"""
data generator for feeding data into pytorch models

NOTE
----
In order to avoid potential error in the methods of slicing signals and rr intervals,
one can check using the following code

.. code-block:: python

    from torch_ecg.databases.datasets.cpsc2021 import CPSC2021Dataset, CPSC2021TrainCfg

    ds_train = CPSC2021Dataset(CPSC2021TrainCfg, task="qrs_detection", training=True)
    ds_val = CPSC2021Dataset(CPSC2021TrainCfg, task="qrs_detection", training=False)
    err_list = []
    for idx, seg in enumerate(ds_train.segments):
        sig, lb = ds_train[idx]
        if sig.shape != (2,6000) or lb.shape != (750, 1):
            print("\n"+f"segment {seg} has sig.shape = {sig.shape}, lb.shape = {lb.shape}"+"\n")
            err_list.append(seg)
        print(f"{idx+1}/{len(ds_train)}", end="\r")
    for idx, seg in enumerate(ds_val.segments):
        sig, lb = ds_val[idx]
        if sig.shape != (2,6000) or lb.shape != (750, 1):
            print("\n"+f"segment {seg} has sig.shape = {sig.shape}, lb.shape = {lb.shape}"+"\n")
            err_list.append(seg)
        print(f"{idx+1}/{len(ds_val)}", end="\r")
    for idx, seg in enumerate(err_list):
        path = ds_train._get_seg_data_path(seg)
        os.remove(path)
        path = ds_train._get_seg_ann_path(seg)
        os.remove(path)
        print(f"{idx+1}/{len(err_list)}", end="\r")

and similarly for the task of `rr_lstm`

"""

import json
import os
import re
import time
import warnings
from copy import deepcopy
from pathlib import Path
from typing import Dict, List, Optional, Sequence, Tuple, Union, Any

import numpy as np
from scipy import signal as SS
from scipy.io import loadmat, savemat
from torch.utils.data.dataset import Dataset
from tqdm.auto import tqdm

from ...._preprocessors import PreprocManager
from ....databases import CPSC2021 as CR
from ....utils.misc import (
    ReprMixin,
    get_record_list_recursive3,
    list_sum,
    nildent,
    add_docstring,
)
from ....utils.utils_data import mask_to_intervals
from ....utils.utils_signal import remove_spikes_naive
from ....utils.utils_data import generate_weight_mask
from ....cfg import CFG, DEFAULTS
from .cpsc2021_cfg import CPSC2021TrainCfg


__all__ = [
    "CPSC2021Dataset",
]


class CPSC2021Dataset(ReprMixin, Dataset):
    """Data generator for feeding data into pytorch models
    using the :class:`~torch_ecg.databases.CPSC2021` database.

    Strategies for generating data and labels:
    1. ECGs are preprocessed and stored in one folder
    2. preprocessed ECGs are sliced with overlap to generate data and label for different tasks:

       - the data files stores segments of fixed length of preprocessed ECGs,
       - the annotation files contain "qrs_mask", and "af_mask"

    The returned values (tuple) of :meth:`__getitem__` depends on the task:

        1. "qrs_detection": (`data`, `qrs_mask`, None)
        2. "rr_lstm": (`rr_seq`, `rr_af_mask`, `rr_weight_mask`)
        3. "main": (`data`, `af_mask`, `weight_mask`)

    where

        - `data` shape: ``(n_lead, n_sample)``
        - `qrs_mask` shape: ``(n_sample, 1)``
        - `af_mask` shape: ``(n_sample, 1)``
        - `weight_mask` shape: ``(n_sample, 1)``
        - `rr_seq` shape: ``(n_rr, 1)``
        - `rr_af_mask` shape: ``(n_rr, 1)``
        - `rr_weight_mask` shape: ``(n_rr, 1)``

    Typical values of ``n_sample`` and ``n_rr`` are 6000 and 30, respectively.

    ``n_lead`` is typically 2, which is the number of leads
    in the ECG signal of the :class:`CPSC2021` database.

    Parameters
    ----------
    config : dict
        Configurations for the dataset, ref. `CPSC2021TrainCfg`.
        A simple example is as follows:

        .. code-block:: python

            >>> config = deepcopy(CPSC2021TrainCfg)
            >>> config.db_dir = "some/path/to/db"
            >>> dataset = CPSC2021Dataset(config, task="main", training=True, lazy=False)

    training : bool, default True
        If True, the training set will be loaded,
        otherwise the test (val) set will be loaded.
    lazy : bool, default True
        If True, the data will not be loaded immediately,
        instead, it will be loaded on demand.
    **reader_kwargs : dict, optional
        Keyword arguments for the database reader class.

    """

    __DEBUG__ = False
    __name__ = "CPSC2021Dataset"

    def __init__(
        self,
        config: CFG,
        task: str,
        training: bool = True,
        lazy: bool = True,
        **reader_kwargs: Any,
    ) -> None:
        super().__init__()
        self.config = deepcopy(config)
        if reader_kwargs.pop("db_dir", None) is not None:
            warnings.warn(
                "`db_dir` is specified in both config and reader_kwargs", RuntimeWarning
            )
        self.reader = CR(db_dir=self.config.db_dir, **reader_kwargs)
        # assert self.config.db_dir is not None, "db_dir must be specified"
        self.config.db_dir = self.reader.db_dir
        self.dtype = self.config.np_dtype
        self.allowed_preproc = list(
            set(
                [
                    "bandpass",
                    "baseline_remove",
                ]
            ).intersection(set(self.config.keys()))
        )

        self.training = training

        self.lazy = lazy

        ppm_config = CFG(random=False)
        ppm_config.update(deepcopy(self.config))
        ppm_config.pop("normalize")
        seg_ppm_config = CFG(random=False)
        seg_ppm_config.update(deepcopy(self.config))
        seg_ppm_config.pop("bandpass")
        self.ppm = PreprocManager.from_config(ppm_config)
        self.seg_ppm = PreprocManager.from_config(seg_ppm_config)

        # create directories if needed
        # preprocess_dir stores pre-processed signals
        self.preprocess_dir = self.config.db_dir / "preprocessed"
        self.preprocess_dir.mkdir(parents=True, exist_ok=True)
        # segments_dir for sliced segments of fixed length
        self.segments_base_dir = self.config.db_dir / "segments"
        self.segments_base_dir.mkdir(parents=True, exist_ok=True)
        self.segment_name_pattern = "S_[\\d]{1,3}_[\\d]{1,2}_[\\d]{7}"
        self.segment_ext = "mat"
        # rr_dir for sequence of rr intervals of fix length
        self.rr_seq_base_dir = self.config.db_dir / "rr_seq"
        self.rr_seq_base_dir.mkdir(parents=True, exist_ok=True)
        self.rr_seq_name_pattern = "R_[\\d]{1,3}_[\\d]{1,2}_[\\d]{7}"
        self.rr_seq_ext = "mat"

        self._all_data = None
        self._all_labels = None
        self._all_masks = None
        self.__set_task(task, lazy=self.lazy)

    def _load_all_data(self) -> None:
        """Load all data into memory."""
        self.__set_task(self.task, lazy=False)

    def __set_task(self, task: str, lazy: bool = True) -> None:
        """Set the task and load the data if needed.

        Parameters
        ----------
        task : str
            Name of the task, can be one of `CPSC2021TrainCfg.tasks`
        lazy : bool, default True
            If True, the data will not be loaded immediately,
            instead, it will be loaded on demand.

        Returns
        -------
        None

        """
        assert task.lower() in CPSC2021TrainCfg.tasks, f"illegal task \042{task}\042"
        if (
            hasattr(self, "task")
            and self.task == task.lower()
            and self._all_data is not None
            and len(self._all_data) > 0
        ):
            return
        self.task = task.lower()
        self.all_classes = self.config[task].classes
        self.n_classes = len(self.config[task].classes)
        self.lazy = lazy

        self.seglen = self.config[task].input_len  # alias, for simplicity
        split_res = self._train_test_split(
            train_ratio=self.config.train_ratio,
            force_recompute=False,
        )
        if self.training:
            self.subjects = split_res.train
        else:
            self.subjects = split_res.test

        if self.task in [
            "qrs_detection",
            "main",
        ]:
            # for qrs detection, or for the main task
            self.segments_dirs = CFG()
            self.__all_segments = CFG()
            self.segments_json = self.segments_base_dir / "segments.json"
            self._ls_segments()
            self.segments = list_sum(
                [self.__all_segments[subject] for subject in self.subjects]
            )
            if self.__DEBUG__:
                self.segments = DEFAULTS.RNG_sample(
                    self.segments, int(len(self.segments) * 0.01)
                ).tolist()
            if self.training:
                DEFAULTS.RNG.shuffle(self.segments)
            # preload data
            self.fdr = _FastDataReader(
                self.config,
                self.task,
                self.seg_ppm,
                self.segments_dirs,
                self.segments,
                self.segment_ext,
            )
            if self.lazy:
                return
            self._all_data, self._all_labels, self._all_masks = [], [], []
            with tqdm(
                range(len(self.fdr)),
                desc="Loading data",
                unit="record",
                dynamic_ncols=True,
                mininterval=1.0,
            ) as pbar:
                for idx in pbar:
                    d, l, m = self.fdr[idx]
                    self._all_data.append(d)
                    self._all_labels.append(l)
                    self._all_masks.append(m)
            self._all_data = np.array(self._all_data).astype(self.dtype)
            self._all_labels = np.array(self._all_labels).astype(self.dtype)
            if self.task == "qrs_detection":
                self._all_masks = None
            else:
                self._all_masks = np.array(self._all_masks).astype(self.dtype)
        elif self.task in [
            "rr_lstm",
        ]:
            self.rr_seq_dirs = CFG()
            self.__all_rr_seq = CFG()
            self.rr_seq_json = self.rr_seq_base_dir / "rr_seq.json"
            self._ls_rr_seq()
            self.rr_seq = list_sum(
                [self.__all_rr_seq[subject] for subject in self.subjects]
            )
            if self.__DEBUG__:
                self.rr_seq = DEFAULTS.RNG_sample(
                    self.rr_seq, int(len(self.rr_seq) * 0.01)
                ).tolist()
            if self.training:
                DEFAULTS.RNG.shuffle(self.rr_seq)
            # preload data
            self.fdr = _FastDataReader(
                self.config,
                self.task,
                self.seg_ppm,
                self.rr_seq_dirs,
                self.rr_seq,
                self.rr_seq_ext,
            )
            if self.lazy:
                return
            self._all_data, self._all_labels, self._all_masks = [], [], []
            with tqdm(
                range(len(self.fdr)),
                desc="Loading data",
                unit="record",
                dynamic_ncols=True,
                mininterval=1.0,
            ) as pbar:
                for idx in pbar:
                    d, l, m = self.fdr[idx]
                    self._all_data.append(d)
                    self._all_labels.append(l)
                    self._all_masks.append(m)
            self._all_data = np.array(self._all_data).astype(self.dtype)
            self._all_labels = np.array(self._all_labels).astype(self.dtype)
            self._all_masks = np.array(self._all_masks).astype(self.dtype)
        else:
            raise NotImplementedError(
                f"data generator for task \042{self.task}\042 not implemented"
            )

    def reset_task(self, task: str, lazy: bool = True) -> None:
        """Reset the task of the data generator.

        Parameters
        ----------
        task : str
            The task to be set.
        lazy : bool, optional
            If True, the data will not be loaded immediately,
            instead, it will be loaded on demand.

        Returns
        -------
        None

        """
        self.__set_task(task, lazy)

    def _ls_segments(self) -> None:
        """Find all segments in the segments directory,
        and store them in some private attributes.
        """
        for item in ["data", "ann"]:
            self.segments_dirs[item] = CFG()
            for s in self.reader.all_subjects:
                self.segments_dirs[item][s] = self.segments_base_dir / item / s
                self.segments_dirs[item][s].mkdir(parents=True, exist_ok=True)
        if self.segments_json.is_file():
            self.__all_segments = json.loads(self.segments_json.read_text())
            # return
        print(
            f"please allow the reader a few minutes to collect the segments from {self.segments_base_dir}..."
        )
        seg_filename_pattern = f"{self.segment_name_pattern}\\.{self.segment_ext}"
        self.__all_segments = CFG(
            {
                s: get_record_list_recursive3(
                    str(self.segments_dirs.data[s]), seg_filename_pattern
                )
                for s in self.reader.all_subjects
            }
        )
        if all([len(self.__all_segments[s]) > 0 for s in self.reader.all_subjects]):
            self.segments_json.write_text(
                json.dumps(self.__all_segments, ensure_ascii=False)
            )

    def _ls_rr_seq(self) -> None:
        """Find all rr sequences in the rr sequences directory,
        and store them in some private attributes.
        """
        for s in self.reader.all_subjects:
            self.rr_seq_dirs[s] = self.rr_seq_base_dir / s
            self.rr_seq_dirs[s].mkdir(parents=True, exist_ok=True)
        if self.rr_seq_json.is_file():
            self.__all_rr_seq = json.loads(self.rr_seq_json.read_text())
            # return
        print(
            f"please allow the reader a few minutes to collect the rr sequences from {self.rr_seq_base_dir}..."
        )
        rr_seq_filename_pattern = f"{self.rr_seq_name_pattern}\\.{self.rr_seq_ext}"
        self.__all_rr_seq = CFG(
            {
                s: get_record_list_recursive3(
                    self.rr_seq_dirs[s], rr_seq_filename_pattern
                )
                for s in self.reader.all_subjects
            }
        )
        if all([len(self.__all_rr_seq[s]) > 0 for s in self.reader.all_subjects]):
            self.rr_seq_json.write_text(
                json.dumps(self.__all_rr_seq, ensure_ascii=False)
            )

    @property
    def all_segments(self) -> CFG:
        if self.task in [
            "qrs_detection",
            "main",
        ]:
            return self.__all_segments
        else:
            return CFG()

    @property
    def all_rr_seq(self) -> CFG:
        if self.task.lower() in [
            "rr_lstm",
        ]:
            return self.__all_rr_seq
        else:
            return CFG()

    def __len__(self) -> int:
        return len(self.fdr)

    def __getitem__(self, index: int) -> Tuple[np.ndarray, ...]:
        if self.lazy:
            if self.task in ["qrs_detection"]:
                return self.fdr[index][:2]
            else:
                return self.fdr[index]
        else:
            if self.task in ["qrs_detection"]:
                return self._all_data[index], self._all_labels[index]
            else:
                return (
                    self._all_data[index],
                    self._all_labels[index],
                    self._all_masks[index],
                )

    def _get_seg_data_path(self, seg: str) -> Path:
        """Get the path of the data file of segment.

        Parameters
        ----------
        seg : str
            Name of the segment, of pattern like "S_1_1_0000193".

        Returns
        -------
        pathlib.Path
            Absolute path of the data file of the segment.

        """
        subject = seg.split("_")[1]
        fp = self.segments_dirs.data[subject] / f"{seg}.{self.segment_ext}"
        return fp

    @add_docstring(_get_seg_data_path.__doc__.replace("data file", "annotation file"))
    def _get_seg_ann_path(self, seg: str) -> Path:
        subject = seg.split("_")[1]
        fp = self.segments_dirs.ann[subject] / f"{seg}.{self.segment_ext}"
        return fp

    def _load_seg_data(self, seg: str) -> np.ndarray:
        """Load the data of the segment.

        Parameters
        ----------
        seg : str
            Name of the segment, of pattern like "S_1_1_0000193".

        Returns
        -------
        numpy.ndarray
            Loaded data of the segment, of shape ``(2, self.seglen)``.

        """
        seg_data_fp = self._get_seg_data_path(seg)
        seg_data = loadmat(str(seg_data_fp))["ecg"]
        return seg_data

    def _load_seg_ann(self, seg: str) -> dict:
        """Load the annotations of the segment.

        Parameters
        ----------
        seg : str
            Name of the segment, of pattern like "S_1_1_0000193".

        Returns
        -------
        seg_ann : dict
            Annotations of the segment, containing:

                - rpeaks: indices of rpeaks of the segment
                - qrs_mask: mask of qrs complexes of the segment
                - af_mask: mask of af episodes of the segment
                - interval: interval ([start_idx, end_idx]) in
                  the original ECG record of the segment

        """
        seg_ann_fp = self._get_seg_ann_path(seg)
        seg_ann = {
            k: v.flatten()
            for k, v in loadmat(str(seg_ann_fp)).items()
            if not k.startswith("__")
        }
        return seg_ann

    def _load_seg_mask(
        self, seg: str, task: Optional[str] = None
    ) -> Union[np.ndarray, Dict[str, np.ndarray]]:
        """Load the mask(s) of segment.

        Parameters
        ----------
        seg : str
            Name of the segment, of pattern like "S_1_1_0000193".
        task : str, optional
            Task of the mask(s) to be loaded, by default the current task.
            If is not None, the current task is overrided;
            if is "all", then all masks ("qrs_mask", "af_mask", etc.)
            will be returned.

        Returns
        -------
        numpy.ndarray or dict
            Mask(s) of the segment,
            of shape ``(self.seglen, self.n_classes)``.

        """
        seg_mask = {
            k: v.reshape((self.seglen, -1))
            for k, v in self._load_seg_ann(seg).items()
            if k
            in [
                "qrs_mask",
                "af_mask",
            ]
        }
        _task = (task or self.task).lower()
        if _task == "all":
            return seg_mask
        if _task in [
            "qrs_detection",
        ]:
            seg_mask = seg_mask["qrs_mask"]
        elif _task in [
            "main",
        ]:
            seg_mask = seg_mask["af_mask"]
        return seg_mask

    def _load_seg_seq_lab(self, seg: str, reduction: int) -> np.ndarray:
        """Load sequence labeling annotations of the segment.

        Parameters
        ----------
        seg : str
            Name of the segment, of pattern like "S_1_1_0000193".
        reduction : int
            Reduction ratio (granularity) of length of the model output,
            compared to the original signal length.

        Returns
        -------
        numpy.ndarray
            Label of the sequence,
            of shape ``(self.seglen//reduction, self.n_classes)``.

        """
        seg_mask = self._load_seg_mask(seg)
        seg_len, n_classes = seg_mask.shape
        seq_lab = np.stack(
            arrays=[
                np.mean(
                    seg_mask[reduction * idx : reduction * (idx + 1)],
                    axis=0,
                    keepdims=True,
                ).astype(int)
                for idx in range(seg_len // reduction)
            ],
            axis=0,
        ).squeeze(axis=1)
        return seq_lab

    def _get_rr_seq_path(self, seq_name: str) -> Path:
        """Get the path of the annotation file of the rr_seq.

        Parameters
        ----------
        seq_name : str
            Name of the rr_seq, of pattern like "R_1_1_0000193".

        Returns
        -------
        pathlib.Path
            absolute path of the annotation file of the rr_seq.

        """
        subject = seq_name.split("_")[1]
        fp = self.rr_seq_dirs[subject] / f"{seq_name}.{self.rr_seq_ext}"
        return fp

    def _load_rr_seq(self, seq_name: str) -> Dict[str, np.ndarray]:
        """Load the metadata of the rr_seq.

        Parameters
        ----------
        seq_name : str
            Name of the rr_seq, of pattern like "R_1_1_0000193".

        Returns
        -------
        dict
            metadata of sequence of rr intervals, including

                - rr: the sequence of rr intervals, with units in seconds,
                  of shape ``(self.seglen, 1)``
                - label: label of the rr intervals, 0 for normal, 1 for af,
                  of shape ``(self.seglen, self.n_classes)``
                - interval: interval of the current rr sequence in the whole
                  rr sequence in the original record

        """
        rr_seq_path = self._get_rr_seq_path(seq_name)
        rr_seq = {
            k: v for k, v in loadmat(str(rr_seq_path)).items() if not k.startswith("__")
        }
        rr_seq["rr"] = rr_seq["rr"].reshape((self.seglen, 1))
        rr_seq["label"] = rr_seq["label"].reshape((self.seglen, self.n_classes))
        rr_seq["interval"] = rr_seq["interval"].flatten()
        return rr_seq

    def persistence(self, force_recompute: bool = False, verbose: int = 0) -> None:
        """Save the preprocessed data to disk.

        Parameters
        ----------
        force_recompute : bool, default False
            Whether to force recompute the preprocessed data.
        verbose : int, default 0
            Verbosity level for printing the progress.

        Returns
        -------
        None

        """
        if verbose >= 1:
            print(" preprocessing data ".center(110, "#"))
        self._preprocess_data(
            force_recompute=force_recompute,
            verbose=verbose,
        )

        original_task = self.task
        original_lazy = self.lazy
        self.__set_task("main", lazy=True)
        if verbose >= 1:
            print("\n" + " slicing data into segments ".center(110, "#"))
        self._slice_data(
            force_recompute=force_recompute,
            verbose=verbose,
        )

        self.__set_task("rr_lstm", lazy=True)
        if verbose >= 1:
            print("\n" + " generating rr sequences ".center(110, "#"))
        self._slice_rr_seq(
            force_recompute=force_recompute,
            verbose=verbose,
        )

        self.__set_task(original_task, lazy=original_lazy)

    def _preprocess_data(self, force_recompute: bool = False, verbose: int = 0) -> None:
        """Preprocesses the ECG data in advance for further use.

        Parameters
        ----------
        force_recompute : bool, default False
            Whether to force recompute the preprocessed data.
        verbose : int, default 0
            Verbosity level for printing the progress.

        Returns
        -------
        None

        """
        for idx, rec in enumerate(self.reader.all_records):
            self._preprocess_one_record(
                rec=rec,
                force_recompute=force_recompute,
                verbose=verbose,
            )
            if verbose >= 1:
                print(f"{idx+1}/{len(self.reader.all_records)} records", end="\r")

    def _preprocess_one_record(
        self, rec: str, force_recompute: bool = False, verbose: int = 0
    ) -> None:
        """Preprocesses one ECG record in advance for further use.

        Parameters
        ----------
        rec : str
            Name of the record.
        force_recompute : bool, default False
            Whether to force recompute the preprocessed data.
        verbose : int, default 0
            Verbosity level for printing the progress.

        Returns
        -------
        None

        """
        suffix = self._get_rec_suffix(self.allowed_preproc)
        save_fp = self.preprocess_dir / f"{rec}-{suffix}.{self.segment_ext}"
        if (not force_recompute) and save_fp.is_file():
            return
        # perform pre-process
        pps, _ = self.ppm(self.reader.load_data(rec), self.config.fs)
        savemat(save_fp, {"ecg": pps}, format="5")

    def load_preprocessed_data(self, rec: str) -> np.ndarray:
        """Load the preprocessed data of the record.

        Parameters
        ----------
        rec : str
            Name of the record.

        Returns
        -------
        numpy.ndarray
            The pre-computed preprocessed ECG data of the record.

        """
        preproc = self.allowed_preproc
        suffix = self._get_rec_suffix(preproc)
        fp = self.preprocess_dir / f"{rec}-{suffix}.{self.segment_ext}"
        if not fp.is_file():
            raise FileNotFoundError(
                f"preprocess(es) \042{preproc}\042 not done for {rec} yet"
            )
        p_sig = loadmat(str(fp))["ecg"]
        if p_sig.shape[0] != 2:
            p_sig = p_sig.T
        return p_sig

    def _get_rec_suffix(self, operations: List[str]) -> str:
        """Get the suffix of the filename of the preprocessed ECG signal.

        Parameters
        ----------
        operations : List[str]
            Names of operations to perform (or has performed).
            Should be sublist of `self.allowed_preproc`.

        Returns
        -------
        str
            Suffix of the filename of the preprocessed ECG signal.

        """
        suffix = "-".join(sorted([item.lower() for item in operations]))
        return suffix

    def _slice_data(self, force_recompute: bool = False, verbose: int = 0) -> None:
        """Slice the preprocessed data into segments.

        Slice all records into segments of length `self.seglen`,
        and perform data augmentations specified in `self.config`.

        Parameters
        ----------
        force_recompute : bool, default False
            Whether to force recompute the preprocessed data.
        verbose : int, default 0
            Verbosity level for printing the progress.

        Returns
        -------
        None

        """
        self.__assert_task(
            [
                "qrs_detection",
                "main",
            ]
        )
        if force_recompute:
            self._clear_cached_segments()
        for idx, rec in enumerate(self.reader.all_records):
            self._slice_one_record(
                rec=rec,
                force_recompute=False,
                update_segments_json=False,
                verbose=verbose,
            )
            if verbose >= 1:
                print(f"{idx+1}/{len(self.reader.all_records)} records", end="\r")
        if force_recompute:
            self.segments_json.write_text(
                json.dump(self.__all_segments, ensure_ascii=False)
            )

    def _slice_one_record(
        self,
        rec: str,
        force_recompute: bool = False,
        update_segments_json: bool = False,
        verbose: int = 0,
    ) -> None:
        """Slice one preprocessed record into segments.

        slice one record into segments of length `self.seglen`,
        and perform data augmentations specified in `self.config`.

        Parameters
        ----------
        rec : str
            Name of the record.
        force_recompute : bool, default False
            Whether to force recompute the preprocessed data.
        update_segments_json : bool, default False
            If both `force_recompute` and `update_segments_json` are True,
            the file `self.segments_json` will be updated.
            Useful when slicing not all records.
        verbose : int, default 0
            Verbosity level for printing the progress.

        Returns
        -------
        None

        """
        self.__assert_task(
            [
                "qrs_detection",
                "main",
            ]
        )
        subject = self.reader.get_subject_id(rec)
        rec_segs = [
            item
            for item in self.__all_segments[subject]
            if item.startswith(rec.replace("data", "S"))
        ]
        if (not force_recompute) and len(rec_segs) > 0:
            return
        elif force_recompute:
            self._clear_cached_segments([rec])

        # data = self.reader.load_data(rec, units="mV")
        data = self.load_preprocessed_data(rec)
        siglen = data.shape[1]
        rpeaks = self.reader.load_rpeaks(rec)
        af_mask = self.reader.load_af_episodes(rec, fmt="mask")
        forward_len = self.seglen - self.config[self.task].overlap_len
        critical_forward_len = self.seglen - self.config[self.task].critical_overlap_len
        critical_forward_len = [critical_forward_len // 4, critical_forward_len]

        # skip those records that are too short
        if siglen < self.seglen:
            return

        # find critical points
        critical_points = np.where(np.diff(af_mask) != 0)[0]
        critical_points = [
            p
            for p in critical_points
            if critical_forward_len[1] <= p < siglen - critical_forward_len[1]
        ]

        segments = []

        # ordinary segments with constant forward_len
        for idx in range((siglen - self.seglen) // forward_len + 1):
            start_idx = idx * forward_len
            new_seg = self.__generate_segment(
                rec=rec,
                data=data,
                start_idx=start_idx,
            )
            segments.append(new_seg)
        # the tail segment
        new_seg = self.__generate_segment(
            rec=rec,
            data=data,
            end_idx=siglen,
        )
        segments.append(new_seg)

        # special segments around critical_points with random forward_len in critical_forward_len
        for cp in critical_points:
            start_idx = max(
                0,
                cp
                - self.seglen
                + DEFAULTS.RNG_randint(
                    critical_forward_len[0], critical_forward_len[1]
                ),
            )
            while start_idx <= min(cp - critical_forward_len[1], siglen - self.seglen):
                new_seg = self.__generate_segment(
                    rec=rec,
                    data=data,
                    start_idx=start_idx,
                )
                segments.append(new_seg)
                start_idx += DEFAULTS.RNG_randint(
                    critical_forward_len[0], critical_forward_len[1]
                )

        # return segments
        self.__save_segments(rec, segments, update_segments_json)

    def __generate_segment(
        self,
        rec: str,
        data: np.ndarray,
        start_idx: Optional[int] = None,
        end_idx: Optional[int] = None,
    ) -> CFG:
        """Generate segment, with possible data augmentation.

        Parameter
        ---------
        rec : str
            Name of the record.
        data : numpy.ndarray
            The whole of (preprocessed) ECG record.
        start_idx : int, optional
            Start index of the signal for generating the segment.
        end_idx : int, optional
            End index of the signal for generating the segment.
            If `start_idx` is set, then `end_idx` is ignored,
            since the segment length is fixed to `self.seglen`.
            At least one of `start_idx` and `end_idx` should be set.

        Returns
        -------
        dict
            Segments (meta-)data, containing:

                - data: values of the segment, with units in mV
                - rpeaks: indices of rpeaks of the segment
                - qrs_mask: mask of qrs complexes of the segment
                - af_mask: mask of af episodes of the segment
                - interval: interval ([start_idx, end_idx]) in the
                  original ECG record of the segment

        """
        assert not all(
            [start_idx is None, end_idx is None]
        ), "at least one of `start_idx` and `end_idx` should be set"
        siglen = data.shape[1]
        # offline augmentations are done, including strech-or-compress, ...
        if self.config.stretch_compress != 0:
            stretch_compress_choices = [0, 1, -1]
            sign = DEFAULTS.RNG_sample(stretch_compress_choices, 1)[0]
            if sign != 0:
                sc_ratio = self.config.stretch_compress
                sc_ratio = (
                    1 + (DEFAULTS.RNG.uniform(sc_ratio / 4, sc_ratio) * sign) / 100
                )
                sc_len = int(round(sc_ratio * self.seglen))
                if start_idx is not None:
                    end_idx = start_idx + sc_len
                else:
                    start_idx = end_idx - sc_len
                if end_idx > siglen:
                    end_idx = siglen
                    start_idx = max(0, end_idx - sc_len)
                    sc_ratio = (end_idx - start_idx) / self.seglen
                aug_seg = data[..., start_idx:end_idx]
                aug_seg = SS.resample(x=aug_seg, num=self.seglen, axis=1)
            else:
                if start_idx is not None:
                    end_idx = start_idx + self.seglen
                    if end_idx > siglen:
                        end_idx = siglen
                        start_idx = end_idx - self.seglen
                else:
                    start_idx = end_idx - self.seglen
                    if start_idx < 0:
                        start_idx = 0
                        end_idx = self.seglen
                # the segment of original signal, with no augmentation
                aug_seg = data[..., start_idx:end_idx]
                sc_ratio = 1
        else:
            if start_idx is not None:
                end_idx = start_idx + self.seglen
                if end_idx > siglen:
                    end_idx = siglen
                    start_idx = end_idx - self.seglen
            else:
                start_idx = end_idx - self.seglen
                if start_idx < 0:
                    start_idx = 0
                    end_idx = self.seglen
            aug_seg = data[..., start_idx:end_idx]
            sc_ratio = 1
        # adjust rpeaks
        seg_rpeaks = self.reader.load_rpeaks(
            rec=rec,
            sampfrom=start_idx,
            sampto=end_idx,
            keep_original=False,
        )
        seg_rpeaks = [
            int(round(r / sc_ratio))
            for r in seg_rpeaks
            if self.config.rpeaks_dist2border
            <= r
            < self.seglen - self.config.rpeaks_dist2border
        ]
        # generate qrs_mask from rpeaks
        seg_qrs_mask = np.zeros((self.seglen,), dtype=int)
        for r in seg_rpeaks:
            seg_qrs_mask[
                r - self.config.qrs_mask_bias : r + self.config.qrs_mask_bias
            ] = 1
        # adjust af_intervals
        seg_af_intervals = self.reader.load_af_episodes(
            rec=rec,
            sampfrom=start_idx,
            sampto=end_idx,
            keep_original=False,
            fmt="intervals",
        )
        seg_af_intervals = [
            [int(round(itv[0] / sc_ratio)), int(round(itv[1] / sc_ratio))]
            for itv in seg_af_intervals
        ]
        # generate af_mask from af_intervals
        seg_af_mask = np.zeros((self.seglen,), dtype=int)
        for itv in seg_af_intervals:
            seg_af_mask[itv[0] : itv[1]] = 1

        new_seg = CFG(
            data=aug_seg,
            rpeaks=seg_rpeaks,
            qrs_mask=seg_qrs_mask,
            af_mask=seg_af_mask,
            interval=[start_idx, end_idx],
        )
        return new_seg

    def __save_segments(
        self, rec: str, segments: List[CFG], update_segments_json: bool = False
    ) -> None:
        """Save the segments to the disk.

        Parameters
        ----------
        rec : str
            Name of the record.
        segments : List[dict]
            List of the segments (meta-)data to be saved.
        update_segments_json : bool, default False
            Whether to update the segments.json file.

        Returns
        -------
        None

        """
        subject = self.reader.get_subject_id(rec)
        ordering = list(range(len(segments)))
        DEFAULTS.RNG.shuffle(ordering)
        for i, idx in enumerate(ordering):
            seg = segments[idx]
            filename = f"{rec}_{i:07d}.{self.segment_ext}".replace("data", "S")
            data_path = self.segments_dirs.data[subject] / filename
            savemat(str(data_path), {"ecg": seg.data})
            self.__all_segments[subject].append(Path(filename).with_suffix("").name)
            ann_path = self.segments_dirs.ann[subject] / filename
            savemat(
                str(ann_path),
                {
                    k: v
                    for k, v in seg.items()
                    if k
                    not in [
                        "data",
                    ]
                },
            )
        if update_segments_json:
            self.segments_json.write_text(
                json.dumps(self.__all_segments, ensure_ascii=False)
            )

    def _clear_cached_segments(self, recs: Optional[Sequence[str]] = None) -> None:
        """Clear the cached segments.

        Parameters
        ----------
        recs : Sequence[str], optional
            Sequence of the records whose segments are to be cleared.
            Defaults to all records.

        Returns
        -------
        None

        """
        self.__assert_task(
            [
                "qrs_detection",
                "main",
            ]
        )
        if recs is not None:
            for rec in recs:
                subject = self.reader.get_subject_id(rec)
                for item in [
                    "data",
                    "ann",
                ]:
                    path = str(self.segments_dirs[item][subject])
                    for f in [
                        n for n in os.listdir(path) if n.endswith(self.segment_ext)
                    ]:
                        if self._get_rec_name(f) == rec:
                            os.remove(os.path.join(path, f))
                            if os.path.splitext(f)[0] in self.__all_segments[subject]:
                                self.__all_segments[subject].remove(
                                    os.path.splitext(f)[0]
                                )
        else:
            for subject in self.reader.all_subjects:
                for item in [
                    "data",
                    "ann",
                ]:
                    path = str(self.segments_dirs[item][subject])
                    for f in [
                        n for n in os.listdir(path) if n.endswith(self.segment_ext)
                    ]:
                        os.remove(os.path.join(path, f))
                        if os.path.splitext(f)[0] in self.__all_segments[subject]:
                            self.__all_segments[subject].remove(os.path.splitext(f)[0])
        self.segments = list_sum(
            [self.__all_segments[subject] for subject in self.subjects]
        )

    def _slice_rr_seq(self, force_recompute: bool = False, verbose: int = 0) -> None:
        """Slice sequences of rr intervals into fixed length (sub)sequences.

        Parameters
        ----------
        force_recompute : bool, default False
            Whether to force recompute the rr sequences.
        verbose : int, default 0
            Verbosity level for printing the progress.

        Returns
        -------
        None

        """
        self.__assert_task(["rr_lstm"])
        if force_recompute:
            self._clear_cached_rr_seq()
        for idx, rec in enumerate(self.reader.all_records):
            self._slice_rr_seq_one_record(
                rec=rec,
                force_recompute=False,
                update_rr_seq_json=False,
                verbose=verbose,
            )
            if verbose >= 1:
                print(f"{idx+1}/{len(self.reader.all_records)} records", end="\r")
        if force_recompute:
            self.rr_seq_json.write_text(
                json.dumps(self.__all_rr_seq, ensure_ascii=False)
            )

    def _slice_rr_seq_one_record(
        self,
        rec: str,
        force_recompute: bool = False,
        update_rr_seq_json: bool = False,
        verbose: int = 0,
    ) -> None:
        """Slice sequences of rr intervals from one record
        into fixed length (sub)sequences.

        Parameters
        ----------
        rec : str
            Name of the record.
        force_recompute : bool, default False
            Whether to force recompute the rr sequences.
        update_rr_seq_json : bool, default False
            Whether to update the rr_seq.json file.
        verbose : int, default 0
            Verbosity level for printing the progress.

        Returns
        -------
        None

        """
        self.__assert_task(["rr_lstm"])
        subject = self.reader.get_subject_id(rec)
        rec_rr_seq = [
            item
            for item in self.__all_rr_seq[subject]
            if item.startswith(rec.replace("data", "R"))
        ]
        if (not force_recompute) and len(rec_rr_seq) > 0:
            return
        elif force_recompute:
            self._clear_cached_rr_seq([rec])

        forward_len = self.seglen - self.config[self.task].overlap_len
        critical_forward_len = self.seglen - self.config[self.task].critical_overlap_len
        critical_forward_len = [critical_forward_len - 2, critical_forward_len]

        rpeaks = self.reader.load_rpeaks(rec)
        rr = np.diff(rpeaks) / self.config.fs
        if len(rr) < self.seglen:
            return
        af_mask = self.reader.load_af_episodes(rec, fmt="mask")
        label_seq = af_mask[rpeaks][:-1]

        # find critical points
        critical_points = np.where(np.diff(label_seq) != 0)[0]
        critical_points = [
            p
            for p in critical_points
            if critical_forward_len[1] <= p < len(rr) - critical_forward_len[1]
        ]

        rr_seq = []

        # ordinary segments with constant forward_len
        for idx in range((len(rr) - self.seglen) // forward_len + 1):
            start_idx = idx * forward_len
            end_idx = start_idx + self.seglen
            new_rr_seq = CFG(
                rr=rr[start_idx:end_idx],
                label=label_seq[start_idx:end_idx],
                interval=[start_idx, end_idx],
            )
            rr_seq.append(new_rr_seq)
        # the tail segment
        if end_idx < len(rr):
            end_idx = len(rr)
            start_idx = end_idx - self.seglen
            new_rr_seq = CFG(
                rr=rr[start_idx:end_idx],
                label=label_seq[start_idx:end_idx],
                interval=[start_idx, end_idx],
            )
            rr_seq.append(new_rr_seq)

        # special segments around critical_points with random forward_len in critical_forward_len
        for cp in critical_points:
            start_idx = max(
                0,
                cp
                - self.seglen
                + DEFAULTS.RNG_randint(
                    critical_forward_len[0], critical_forward_len[1]
                ),
            )
            while start_idx <= min(cp - critical_forward_len[1], len(rr) - self.seglen):
                end_idx = start_idx + self.seglen
                new_rr_seq = CFG(
                    rr=rr[start_idx:end_idx],
                    label=label_seq[start_idx:end_idx],
                    interval=[start_idx, end_idx],
                )
                rr_seq.append(new_rr_seq)
                start_idx += DEFAULTS.RNG_randint(
                    critical_forward_len[0], critical_forward_len[1]
                )
        # save rr sequences
        self.__save_rr_seq(rec, rr_seq, update_rr_seq_json)

    def __save_rr_seq(
        self, rec: str, rr_seq: List[CFG], update_rr_seq_json: bool = False
    ) -> None:
        """Save the sliced rr sequences to disk.

        Parameters
        ----------
        rec : str,
            Name of the record.
        rr_seq : List[dict],
            List of the rr_seq (meta-)data.
        update_rr_seq_json : bool, default False
            Whether to update the rr_seq.json file.

        Returns
        -------
        None

        """
        subject = self.reader.get_subject_id(rec)
        ordering = list(range(len(rr_seq)))
        DEFAULTS.RNG.shuffle(ordering)
        for i, idx in enumerate(ordering):
            item = rr_seq[idx]
            filename = f"{rec}_{i:07d}.{self.rr_seq_ext}".replace("data", "R")
            data_path = self.rr_seq_dirs[subject] / filename
            savemat(str(data_path), item)
            self.__all_rr_seq[subject].append(Path(filename).with_suffix("").name)
        if update_rr_seq_json:
            self.rr_seq_json.write_text(
                json.dumps(self.__all_rr_seq, ensure_ascii=False)
            )

    def _clear_cached_rr_seq(self, recs: Optional[Sequence[str]] = None) -> None:
        """Clear the cached rr sequences.

        Parameters
        ----------
        recs : Sequence[str], optional
            Sequence of the records whose cached rr sequences are to be cleared.
            Defaults to all records.

        Returns
        -------
        None

        """
        self.__assert_task(["rr_lstm"])
        if recs is not None:
            for rec in recs:
                subject = self.reader.get_subject_id(rec)
                path = str(self.rr_seq_dirs[subject])
                for f in [n for n in os.listdir(path) if n.endswith(self.rr_seq_ext)]:
                    if self._get_rec_name(f) == rec:
                        os.remove(os.path.join(path, f))
                        if os.path.splitext(f)[0] in self.__all_rr_seq[subject]:
                            self.__all_rr_seq[subject].remove(os.path.splitext(f)[0])
        else:
            for subject in self.reader.all_subjects:
                path = str(self.rr_seq_dirs[subject])
                for f in [n for n in os.listdir(path) if n.endswith(self.rr_seq_ext)]:
                    os.remove(os.path.join(path, f))
                    if os.path.splitext(f)[0] in self.__all_rr_seq[subject]:
                        self.__all_rr_seq[subject].remove(os.path.splitext(f)[0])
        self.rr_seq = list_sum(
            [self.__all_rr_seq[subject] for subject in self.subjects]
        )

    def _get_rec_name(self, seg_or_rr: str) -> str:
        """Get the name of the record that a segment
        or rr_seq was generated from.

        Parameters
        ----------
        seg_or_rr : str
            Name of the segment or rr_seq.

        Returns
        -------
        rec : str
            Name of the record
            that the segment or rr_seq was generated from.

        """
        rec = re.sub("[RS]", "data", os.path.splitext(seg_or_rr)[0])[:-8]
        return rec

    def _train_test_split(
        self, train_ratio: float = 0.8, force_recompute: bool = False
    ) -> Dict[str, List[str]]:
        """Perform train-test split.

        Parameters
        ----------
        train_ratio : float, default 0.8
            Ratio of the train set in the whole dataset.
        force_recompute : bool, default False
            If True, the train-test split will be recomputed,
            regardless of the existing ones stored in json files.

        Returns
        -------
        dict
            Dictionary of the train-test split.
            Keys are "train" and "test", and
            values are list of the subjects split for training or validation.

        """
        start = time.time()
        print("\nstart performing train test split...\n")
        _train_ratio = int(train_ratio * 100)
        _test_ratio = 100 - _train_ratio
        assert _train_ratio * _test_ratio > 0

        train_file = self.reader.db_dir_base / f"train_ratio_{_train_ratio}.json"
        test_file = self.reader.db_dir_base / f"test_ratio_{_test_ratio}.json"

        if force_recompute or not all([train_file.is_file(), test_file.is_file()]):
            all_subjects = set(self.reader.df_stats.subject_id.tolist())
            afp_subjects = set(
                self.reader.df_stats[
                    self.reader.df_stats.label == "AFp"
                ].subject_id.tolist()
            )
            aff_subjects = (
                set(
                    self.reader.df_stats[
                        self.reader.df_stats.label == "AFf"
                    ].subject_id.tolist()
                )
                - afp_subjects
            )
            normal_subjects = all_subjects - afp_subjects - aff_subjects

            test_set = (
                DEFAULTS.RNG_sample(
                    list(afp_subjects),
                    max(1, int(round(len(afp_subjects) * _test_ratio / 100))),
                ).tolist()
                + DEFAULTS.RNG_sample(
                    list(aff_subjects),
                    max(1, int(round(len(aff_subjects) * _test_ratio / 100))),
                ).tolist()
                + DEFAULTS.RNG_sample(
                    list(normal_subjects),
                    max(1, int(round(len(normal_subjects) * _test_ratio / 100))),
                ).tolist()
            )
            train_set = list(all_subjects - set(test_set))

            DEFAULTS.RNG.shuffle(test_set)
            DEFAULTS.RNG.shuffle(train_set)

            train_file.write_text(json.dumps(train_set, ensure_ascii=False))
            test_file.write_text(json.dumps(test_set, ensure_ascii=False))
            print(
                nildent(
                    f"""
                train set saved to \n\042{str(train_file)}\042
                test set saved to \n\042{str(test_file)}\042
                """
                )
            )
        else:
            train_set = json.loads(train_file.read_text())
            test_set = json.loads(test_file.read_text())

        print(f"train test split finished in {(time.time()-start)/60:.2f} minutes")

        split_res = CFG(
            {
                "train": train_set,
                "test": test_set,
            }
        )
        return split_res

    def __assert_task(self, tasks: List[str]) -> None:
        """Check if the current task is in the given list of tasks."""
        assert self.task in tasks, (
            f"DO NOT call this method when the current task is {self.task}. "
            "Switch task using `reset_task`"
        )

    def plot_seg(self, seg: str, ticks_granularity: int = 0) -> None:
        """Plot the segment.

        Parameters
        ----------
        seg : str
            Name of the segment, of pattern like "S_1_1_0000193".
        ticks_granularity : int, default 0
            Granularity to plot axis ticks, the higher the more ticks.
            0 (no ticks) --> 1 (major ticks) --> 2 (major + minor ticks)

        Returns
        -------
        None

        """
        seg_data = self._load_seg_data(seg)
        print(f"seg_data.shape = {seg_data.shape}")
        seg_ann = self._load_seg_ann(seg)
        seg_ann["af_episodes"] = mask_to_intervals(seg_ann["af_mask"], vals=1)
        print(f"seg_ann = {seg_ann}")
        rec_name = self._get_rec_name(seg)
        self.reader.plot(
            rec=rec_name,  # unnecessary indeed
            data=seg_data,
            ann=seg_ann,
            ticks_granularity=ticks_granularity,
        )

    def extra_repr_keys(self) -> List[str]:
        return [
            "training",
            "task",
            "reader",
        ]


class _FastDataReader(ReprMixin, Dataset):
    """Fast data reader.

    Parameters
    ----------
    reader : CR
        The reader to read the data.
    records : Sequence[str]
        The list of records to read.
    config : CFG
        The configuration.
    ppm : PreprocManager, optional
        The preprocessor manager.

    """

    def __init__(
        self,
        config: CFG,
        task: str,
        seg_ppm: PreprocManager,
        file_dirs: dict,
        files: List[str],
        file_ext: str,
    ) -> None:
        self.config = config
        self.task = task
        self.seg_ppm = seg_ppm
        self.file_dirs = file_dirs
        self.files = files
        self.file_ext = file_ext

        self.seglen = self.config[self.task].input_len
        self.n_classes = len(self.config[task].classes)

        self._seg_keys = {
            "qrs_detection": "qrs_mask",
            "main": "af_mask",
        }

    def __getitem__(self, index: int) -> Tuple[np.ndarray, ...]:
        if self.task in [
            "qrs_detection",
            "main",
        ]:
            seg_name = self.files[index]
            subject = seg_name.split("_")[1]
            seg_data_fp = self.file_dirs.data[subject] / f"{seg_name}.{self.file_ext}"
            seg_data = loadmat(str(seg_data_fp))["ecg"]
            for idx in range(seg_data.shape[0]):
                seg_data[idx] = remove_spikes_naive(seg_data[idx])
            seg_ann_fp = self.file_dirs.ann[subject] / f"{seg_name}.{self.file_ext}"
            seg_label = loadmat(str(seg_ann_fp))[self._seg_keys[self.task]].reshape(
                (self.seglen, -1)
            )
            if self.config[self.task].reduction > 1:
                reduction = self.config[self.task].reduction
                seg_len, n_classes = seg_label.shape
                seg_label = np.stack(
                    arrays=[
                        np.mean(
                            seg_data[reduction * idx : reduction * (idx + 1)],
                            axis=0,
                            keepdims=True,
                        ).astype(int)
                        for idx in range(seg_len // reduction)
                    ],
                    axis=0,
                ).squeeze(axis=1)
            seg_data, _ = self.seg_ppm(seg_data, self.config.fs)
            if self.task == "main":
                weight_mask = generate_weight_mask(
                    target_mask=seg_label.squeeze(-1),
                    fg_weight=2,
                    fs=self.config.fs,
                    reduction=self.config[self.task].reduction,
                    radius=0.8,
                    boundary_weight=5,
                )[..., np.newaxis]
                return seg_data, seg_label, weight_mask
            return seg_data, seg_label, None
        elif self.task in [
            "rr_lstm",
        ]:
            seq_name = self.files[index]
            subject = seq_name.split("_")[1]
            rr_seq_path = self.file_dirs[subject] / f"{seq_name}.{self.file_ext}"
            rr_seq = loadmat(str(rr_seq_path))
            rr_seq["rr"] = rr_seq["rr"].reshape((self.seglen, 1))
            rr_seq["label"] = rr_seq["label"].reshape((self.seglen, self.n_classes))
            weight_mask = generate_weight_mask(
                target_mask=rr_seq["label"].squeeze(-1),
                fg_weight=2,
                fs=1 / 0.8,
                reduction=1,
                radius=2,
                boundary_weight=5,
            )[..., np.newaxis]
            return rr_seq["rr"], rr_seq["label"], weight_mask
        else:
            raise NotImplementedError(
                f"data generator for task \042{self.task}\042 not implemented"
            )

    def __len__(self) -> int:
        return len(self.files)

    def extra_repr_keys(self) -> List[str]:
        return [
            "task",
            "reader",
            "ppm",
        ]


# `StandaloneSegmentSlicer` can be found in `benchmarks/train_hybrid_cpsc2021/dataset.py`



"""
Designing Network Design Spaces
"""

import math
import textwrap
import warnings
from collections import Counter
from itertools import repeat
from numbers import Real
from typing import Optional, Sequence, Union, List

import torch
from torch import nn

from ...cfg import CFG
from ...models._nets import (
    Conv_Bn_Activation,
    DownSample,
    SpaceToDepth,
)
from ...utils.misc import add_docstring, CitationMixin
from ...utils.utils_nn import (
    SizeMixin,
    compute_sequential_output_shape,
    compute_sequential_output_shape_docstring,
)
from .resnet import ResNetBottleNeck, ResNetBasicBlock


class AnyStage(nn.Sequential, SizeMixin):
    """AnyStage of :class:`RegNet`.

    Parameters
    ----------
    in_channels : int
        Number of features (channels) of the input.
    num_filters : Sequence[int]
        Number of filters for the neck conv layer.
    filter_length : int
        Lengths (sizes) of the filter kernels for the neck conv layer.
    subsample_length : int
        Subsample length, including pool size for short cut,
        and stride for the (top or neck) conv layer.
    num_blocks : int
        Number of blocks in the stage.
    group_width : int
        Group width for the bottleneck block.
    stage_index : int
        Index of the stage in the whole :class:`RegNet`.
    block_config: dict,
        (optional) configs for the blocks, including
            - block: str or torch.nn.Module,
              the block class, can be one of
              "bottleneck", "bottle_neck", :class:`ResNetBottleNeck`, etc.
            - expansion: int,
              the expansion factor for the bottleneck block.
            - increase_channels_method: str,
              the method to increase the number of channels,
              can be one of {"conv", "zero_padding"}.
            - subsample_mode: str,
              the mode of subsampling, can be one of
              {:class:`DownSample`.__MODES__},
            - activation: str or torch.nn.Module,
              the activation function, can be one of
              {:class:`Activations`}.
            - kw_activation: dict,
              keyword arguments for the activation function.
            - kernel_initializer: str,
              the kernel initializer, can be one of
              {:class:`Initializers`}.
            - kw_initializer: dict,
              keyword arguments for the kernel initializer.
            - bias: bool,
              whether to use bias in the convolution.
            - dilation: int,
              the dilation factor for the convolution.
            - base_width: int,
              number of filters per group for the neck conv layer
              usually number of filters of the initial conv layer
              of the whole :class:`RegNet`.
            - base_groups: int,
              pattern of connections between inputs and outputs of
              conv layers at the two ends, which should divide `groups`.
            - base_filter_length: int,
              lengths (sizes) of the filter kernels for conv layers at the two ends.
            - attn: dict,
              attention mechanism for the neck conv layer.
              If is None, no attention mechanism is used.
              If is not None, it should be a dict with the following items:

                - name: str, can be "se", "gc", "nl" (alias "nonlocal", "non-local"), etc.
                - pos: int, position of the attention mechanism.

              Other keys are specific to the attention mechanism.

    """

    __name__ = "AnyStage"
    __DEFAULT_BLOCK_CONFIG__ = {
        "block": "bottleneck",
        "expansion": 1,
        "increase_channels_method": "conv",
        "subsample_mode": "conv",
        "activation": "relu",
        "kw_activation": {"inplace": True},
        "kernel_initializer": "he_normal",
        "kw_initializer": {},
        "bias": False,
    }

    def __init__(
        self,
        in_channels: int,
        num_filters: int,
        filter_length: int,
        subsample_length: int,
        num_blocks: int,
        group_width: int,
        stage_index: int,
        **block_config,
    ) -> None:
        super().__init__()

        self.block_config = CFG(self.__DEFAULT_BLOCK_CONFIG__.copy())
        self.block_config.update(block_config)

        block_cls = self.get_building_block_cls(self.block_config)

        # adjust num_filters based on group_width
        if num_filters % group_width != 0:
            _num_filters = num_filters // group_width * group_width
            if _num_filters < 0.9 * num_filters:
                _num_filters += group_width
            num_filters = _num_filters
        groups = num_filters // group_width
        base_width = block_cls.__DEFAULT_BASE_WIDTH__ / groups

        block_in_channels = in_channels
        for i in range(num_blocks):
            block = block_cls(
                in_channels=block_in_channels,
                num_filters=num_filters,
                filter_length=filter_length,
                subsample_length=subsample_length if i == 0 else 1,
                groups=groups,
                base_width=base_width,
                **self.block_config,
            )
            block_in_channels = block.compute_output_shape()[1]
            self.add_module(f"block_{stage_index}_{i}", block)

    @staticmethod
    def get_building_block_cls(config: CFG) -> nn.Module:
        """Get the building block class."""
        block_cls = config.get("block")
        if isinstance(block_cls, str):
            if block_cls.lower() in ["bottleneck", "bottle_neck"]:
                block_cls = ResNetBottleNeck
            else:
                block_cls = ResNetBasicBlock
        return block_cls

    @add_docstring(
        textwrap.indent(compute_sequential_output_shape_docstring, " " * 4),
        mode="append",
    )
    def compute_output_shape(
        self, seq_len: Optional[int] = None, batch_size: Optional[int] = None
    ) -> Sequence[Union[int, None]]:
        """Compute the output shape of the stage."""
        return compute_sequential_output_shape(self, seq_len, batch_size)


class RegNetStem(nn.Sequential, SizeMixin):
    """The input stem of :class:`RegNet`.

    Parameters
    ----------
    in_channels : int
        Number of input channels.
    out_channels: int or Sequence[int]
        Number of output channels.
    filter_lengths : int or Sequence[int]
        Length of the filter, or equivalently,
        the kernel size(s) of the convolutions.
    conv_stride : int
        Stride of the convolution.
    pool_size : int
        Size of the pooling window.
    pool_stride : int
        Stride of the pooling window.
    subsample_mode : str
        Mode of subsampling, can be one of
        {:class:`DownSample`.__MODES__},
        or "s2d" (with aliases "space_to_depth", "SpaceToDepth").
    groups : int
        Number of groups for the convolution.
    config : dict
        Other configs for convolution and pooling.

    """

    __name__ = "ResNetStem"

    def __init__(
        self,
        in_channels: int,
        out_channels: Union[int, Sequence[int]],
        filter_lengths: Union[int, Sequence[int]],
        conv_stride: int,
        pool_size: int,
        pool_stride: int,
        subsample_mode: str = "max",
        groups: int = 1,
        **config,
    ) -> None:
        super().__init__()
        self.__in_channels = in_channels
        self.__out_channels = out_channels
        self.__filter_lengths = filter_lengths
        if subsample_mode.lower() in ["s2d", "space_to_depth", "SpaceToDepth"]:
            self.add_module(
                "s2d",
                SpaceToDepth(
                    self.__in_channels, self.__out_channels, config.get("block_size", 4)
                ),
            )
            return
        if isinstance(self.__filter_lengths, int):
            self.__filter_lengths = [self.__filter_lengths]
        if isinstance(self.__out_channels, int):
            self.__out_channels = [self.__out_channels]
        assert len(self.__filter_lengths) == len(self.__out_channels)

        conv_in_channels = self.__in_channels
        for idx, fl in enumerate(self.__filter_lengths):
            self.add_module(
                f"conv_{idx}",
                Conv_Bn_Activation(
                    conv_in_channels,
                    self.__out_channels[idx],
                    self.__filter_lengths[idx],
                    stride=conv_stride if idx == 0 else 1,
                    groups=groups,
                    **config,
                ),
            )
            conv_in_channels = self.__out_channels[idx]
        if pool_stride > 1:
            self.add_module(
                "pool",
                DownSample(
                    pool_stride,
                    conv_in_channels,
                    kernel_size=pool_size,
                    groups=groups,
                    padding=(pool_stride - 1) // 2,
                    mode=subsample_mode.lower(),
                    **config,
                ),
            )

    @add_docstring(
        textwrap.indent(compute_sequential_output_shape_docstring, " " * 4),
        mode="append",
    )
    def compute_output_shape(
        self, seq_len: Optional[int] = None, batch_size: Optional[int] = None
    ) -> Sequence[Union[int, None]]:
        """Compute the output shape of the stem."""
        return compute_sequential_output_shape(self, seq_len, batch_size)


class RegNet(nn.Sequential, SizeMixin, CitationMixin):
    """RegNet model.

    RegNet is a family of convolutional neural networks that can be
    constructed by efficiently scaling and pruning a single
    convolutional "stem" network. This architecture is proposed in
    [1]_, and the implementation is adapted from [2]_.

    References
    ----------
    .. [1] https://arxiv.org/abs/2003.13678
    .. [2] https://github.com/pytorch/vision/blob/master/torchvision/models/regnet.py

    Parameters
    ----------
    in_channels : int
        Number of channels of the input.
    config : dict
        Hyper-parameters of the Module, ref. corr. config file.
        Keyword arguments that must be set:

            - filter_lengths: int or sequence of int,
              filter length(s) (kernel size(s)) of the convolutions,
              with granularity to the whole network, to each stage.
            - subsample_lengths: int or sequence of int,
              subsampling length(s) (ratio(s)) of all blocks,
              with granularity to the whole network, to each stage.
            - tot_blocks: int,
              the total number of building blocks.
            - w_a, w_0, w_m: float,
              the parameters for the widths generating function.
            - group_widths: int or sequence of int,
              the number of channels in each group,
              with granularity to the whole network, to each stage.
            - num_blocks: sequence of int, optional,
              the number of blocks in each stage,
              if not given, will be computed from tot_blocks
              and `w_a`, `w_0`, `w_m`.
            - num_filters: int or sequence of int, optional,
              the number of filters in each stage.
              If not given, will be computed from tot_blocks
              and `w_a`, `w_0`, `w_m`.
            - stem: dict,
              the config of the input stem.
            - block: dict,
              other parameters that can be set for the building blocks.

    """

    __name__ = "RegNet"
    __DEFAULT_CONFIG__ = dict(
        activation="relu",
        kw_activation={"inplace": True},
        kernel_initializer="he_normal",
        kw_initializer={},
        base_groups=1,
        dropouts=0,
    )

    def __init__(self, in_channels: int, **config) -> None:
        super().__init__()
        self.__in_channels = in_channels
        self.config = CFG(self.__DEFAULT_CONFIG__.copy())
        self.config.update(config)

        stem_config = CFG(self.config.stem)
        stem_config.pop("num_filters", None)
        self.add_module(
            "input_stem",
            RegNetStem(
                in_channels=self.__in_channels,
                out_channels=self.config.stem.num_filters,
                groups=self.config.base_groups,
                activation=self.config.activation,
                **stem_config,
            ),
        )

        stage_configs = self._get_stage_configs()
        in_channels = self.input_stem.compute_output_shape()[1]
        for idx, stage_config in enumerate(stage_configs):
            stage_block = AnyStage(in_channels=in_channels, **stage_config)
            self.add_module(f"stage_{idx}", stage_block)
            in_channels = stage_block.compute_output_shape()[1]

    def _get_stage_configs(self) -> List[CFG]:
        """Get the configs for each stage."""
        stage_configs = []
        if self.config.get("num_blocks", None) is not None:
            if isinstance(self.config.filter_lengths, int):
                self.__filter_lengths = list(
                    repeat(self.config.filter_lengths, len(self.config.num_blocks))
                )
            else:
                self.__filter_lengths = self.config.filter_lengths
            assert len(self.__filter_lengths) == len(self.config.num_blocks), (
                f"`config.filter_lengths` indicates {len(self.__filter_lengths)} stages, "
                f"while `config.num_blocks` indicates {len(self.config.num_blocks)}"
            )
            if isinstance(self.config.subsample_lengths, int):
                self.__subsample_lengths = list(
                    repeat(self.config.subsample_lengths, len(self.config.num_blocks))
                )
            else:
                self.__subsample_lengths = self.config.subsample_lengths
            assert len(self.__subsample_lengths) == len(self.config.num_blocks), (
                f"`config.subsample_lengths` indicates {len(self.__subsample_lengths)} stages, "
                f"while `config.num_blocks` indicates {len(self.config.num_blocks)}"
            )
            self.__num_filters = self.config.num_filters
            assert len(self.__num_filters) == len(self.config.num_blocks), (
                f"`config.num_filters` indicates {len(self.__num_filters)} stages, "
                f"while `config.num_blocks` indicates {len(self.config.num_blocks)}"
            )
            if isinstance(self.config.dropouts, Real):
                self.__dropouts = list(
                    repeat(self.config.dropouts, len(self.config.num_blocks))
                )
            else:
                self.__dropouts = self.config.dropouts
            assert len(self.__dropouts) == len(self.config.num_blocks), (
                f"`config.dropouts` indicates {len(self.__dropouts)} stages, "
                f"while `config.num_blocks` indicates {len(self.config.num_blocks)}"
            )
            if isinstance(self.config.group_widths, int):
                self.__group_widths = list(
                    repeat(self.config.group_widths, len(self.config.num_blocks))
                )
            else:
                self.__group_widths = self.config.group_widths
            assert len(self.__group_widths) == len(self.config.num_blocks), (
                f"`config.group_widths` indicates {len(self.__group_widths)} stages, "
                f"while `config.num_blocks` indicates {len(self.config.num_blocks)}"
            )

            block_config = CFG(self.config.get("block", {}))
            block_config.pop("dropout", None)

            stage_configs = [
                CFG(
                    dict(
                        num_blocks=self.config.num_blocks[idx],
                        num_filters=self.__num_filters[idx],
                        filter_length=self.__filter_lengths[idx],
                        subsample_length=self.__subsample_lengths[idx],
                        dropout=self.__dropouts[idx],
                        group_width=self.__group_widths[idx],
                        stage_index=idx,
                        **block_config,
                    )
                )
                for idx in range(len(self.config.num_blocks))
            ]
            return stage_configs

        if self.config.get("num_filters", None) is not None:
            warnings.warn(
                "`num_filters` are computed from `config.w_a`, `config.w_0`, `config.w_m`, "
                "if `config.num_blocks` is not provided. "
                "This may not be the intended behavior.",
                RuntimeWarning,
            )
        assert {"w_a", "w_0", "w_m", "tot_blocks"}.issubset(set(self.config.keys())), (
            "If `config.num_blocks` is not provided, then `config.w_a`, `config.w_0`, `config.w_m`, "
            "and `config.tot_blocks` must be provided."
        )
        QUANT = 8
        if (
            self.config.w_a < 0
            or self.config.w_0 <= 0
            or self.config.w_m <= 1
            or self.config.w_0 % QUANT != 0
        ):
            raise ValueError("Invalid RegNet settings")
        # Compute the block widths. Each stage has one unique block width
        widths_cont = (
            torch.arange(self.config.tot_blocks) * self.config.w_a + self.config.w_0
        )
        block_capacity = torch.round(
            torch.log(widths_cont / self.config.w_0) / math.log(self.config.w_m)
        )
        block_widths = (
            (
                torch.round(
                    torch.divide(
                        self.config.w_0 * torch.pow(self.config.w_m, block_capacity),
                        QUANT,
                    )
                )
                * QUANT
            )
            .int()
            .tolist()
        )
        counter = Counter(block_widths)
        num_stages = len(counter)

        if isinstance(self.config.filter_lengths, int):
            self.__filter_lengths = list(repeat(self.config.filter_lengths, num_stages))
        else:
            self.__filter_lengths = self.config.filter_lengths
        assert len(self.__filter_lengths) == num_stages, (
            f"`config.filter_lengths` indicates {len(self.__filter_lengths)} stages, "
            f"while there are {num_stages} computed from "
            "`config.w_a`, `config.w_0`, `config.w_m`, `config.tot_blocks`"
        )
        if isinstance(self.config.subsample_lengths, int):
            self.__subsample_lengths = list(
                repeat(self.config.subsample_lengths, num_stages)
            )
        else:
            self.__subsample_lengths = self.config.subsample_lengths
        assert len(self.__subsample_lengths) == num_stages, (
            f"`config.subsample_lengths` indicates {len(self.__subsample_lengths)} stages, "
            f"while there are {num_stages} computed from "
            "`config.w_a`, `config.w_0`, `config.w_m`, `config.tot_blocks`"
        )
        if isinstance(self.config.dropouts, Real):
            self.__dropouts = list(repeat(self.config.dropouts, num_stages))
        else:
            self.__dropouts = self.config.dropouts
        assert len(self.__dropouts) == num_stages, (
            f"`config.dropouts` indicates {len(self.__dropouts)} stages, "
            f"while there are {num_stages} computed from "
            "`config.w_a`, `config.w_0`, `config.w_m`, `config.tot_blocks`"
        )
        if isinstance(self.config.group_widths, int):
            self.__group_widths = list(repeat(self.config.group_widths, num_stages))
        else:
            self.__group_widths = self.config.group_widths
        assert len(self.__group_widths) == num_stages, (
            f"`config.group_widths` indicates {len(self.__group_widths)} stages, "
            f"while there are {num_stages} computed from "
            "`config.w_a`, `config.w_0`, `config.w_m`, `config.tot_blocks`"
        )
        block_config = CFG(self.config.get("block", {}))
        block_config.pop("dropout", None)

        for idx, num_filters in enumerate(sorted(counter)):
            stage_configs.append(
                CFG(
                    dict(
                        num_blocks=counter[num_filters],
                        num_filters=num_filters,
                        filter_length=self.__filter_lengths[idx],
                        subsample_length=self.__subsample_lengths[idx],
                        group_width=self.__group_widths[idx],
                        dropout=self.__dropouts[idx],
                        stage_index=idx,
                        **block_config,
                    )
                )
            )
        return stage_configs

    @add_docstring(
        textwrap.indent(compute_sequential_output_shape_docstring, " " * 4),
        mode="append",
    )
    def compute_output_shape(
        self, seq_len: Optional[int] = None, batch_size: Optional[int] = None
    ) -> Sequence[Union[int, None]]:
        """Compute the output shape of the network."""
        return compute_sequential_output_shape(self, seq_len, batch_size)

    @property
    def in_channels(self) -> int:
        return self.__in_channels

    @property
    def doi(self) -> List[str]:
        return list(set(self.config.get("doi", []) + ["10.48550/ARXIV.2003.13678"]))



"""Add baseline wander composed of sinusoidal and Gaussian noise to the ECGs."""

import multiprocessing as mp
from itertools import repeat
from numbers import Real
from random import randint
from typing import Any, List, Optional, Sequence, Tuple, Union

import numpy as np
import torch
from torch import Tensor

from ..cfg import DEFAULTS
from ..utils.utils_signal import get_ampl
from .base import Augmenter


__all__ = [
    "BaselineWanderAugmenter",
]


class BaselineWanderAugmenter(Augmenter):
    """Generate baseline wander composed of
    sinusoidal and Gaussian noise.

    Parameters
    ----------
    fs : int
        Sampling frequency of the ECGs to be augmented
    bw_fs : numpy.ndarray, optional
        Frequencies of the sinusoidal noises,
        of shape ``(n,)``,
        defaults to ``[0.33, 0.1, 0.05, 0.01]``.
    ampl_ratio : numpy.ndarray, optional
        Candidate ratios of noise amplitdes compared to the original ECGs for each `fs`,
        of shape ``(m, n)``,
        defaults to

        .. code-block:: python

            np.array(
                [
                    [0.01, 0.01, 0.02, 0.03],  # low
                    [0.01, 0.02, 0.04, 0.05],  # low
                    [0.1, 0.06, 0.04, 0.02],  # low
                    [0.02, 0.04, 0.07, 0.1],  # low
                    [0.05, 0.1, 0.16, 0.25],  # medium
                    [0.1, 0.15, 0.25, 0.3],  # high
                    [0.25, 0.25, 0.3, 0.35],  # extremely high
                ]
            )

    gaussian : numpy.ndarray, optional
        Candidate mean and std of the Gaussian noises,
        of shape ``(k, 2)``,
        defaults to

        .. code-block:: python

            np.array(
                [  # mean and std, in terms of ratio
                    [0.0, 0.001],
                    [0.0, 0.003],
                    [0.0, 0.01],
                ]
            )

    prob : float, default 0.5
        Probability of performing the augmentation.
    inplace : bool, default True
        If True, ECG signal tensors will be modified inplace.
    kwargs : dict, optional
        Additional keyword arguments.

    Examples
    --------
    .. code-block:: python

        blw = BaselineWanderAugmenter(300, prob=0.7)
        sig = torch.randn(32, 12, 5000)
        label = torch.ones((32, 20))
        sig, _ = blw(sig, label)

    """

    __name__ = "BaselineWanderAugmenter"

    def __init__(
        self,
        fs: int,
        bw_fs: Optional[np.ndarray] = None,
        ampl_ratio: Optional[np.ndarray] = None,
        gaussian: Optional[np.ndarray] = None,
        prob: float = 0.5,
        inplace: bool = True,
        **kwargs: Any
    ) -> None:
        super().__init__()
        self.fs = fs
        self.bw_fs = bw_fs if bw_fs is not None else np.array([0.33, 0.1, 0.05, 0.01])
        self.prob = prob
        assert 0 <= self.prob <= 1, "Probability must be between 0 and 1"
        self.ampl_ratio = (
            ampl_ratio
            if ampl_ratio is not None
            else np.array(
                [  # default ampl_ratio
                    [0.01, 0.01, 0.02, 0.03],  # low
                    [0.01, 0.02, 0.04, 0.05],  # low
                    [0.1, 0.06, 0.04, 0.02],  # low
                    [0.02, 0.04, 0.07, 0.1],  # low
                    [0.05, 0.1, 0.16, 0.25],  # medium
                    [0.1, 0.15, 0.25, 0.3],  # high
                    [0.25, 0.25, 0.3, 0.35],  # extremely high
                ]
            )
        )
        if self.prob > 0:
            self.ampl_ratio = np.concatenate(
                (
                    np.zeros(
                        (
                            int((1 - self.prob) * self.ampl_ratio.shape[0] / self.prob),
                            self.ampl_ratio.shape[1],
                        )
                    ),
                    self.ampl_ratio,
                )
            )
        self.gaussian = (
            gaussian
            if gaussian is not None
            else np.array(
                [  # default gaussian, mean and std, in terms of ratio
                    [0.0, 0.001],
                    [0.0, 0.003],
                    [0.0, 0.01],
                ]
            )
        )
        if self.prob > 0:
            self.gaussian = np.concatenate(
                (
                    np.zeros(
                        (
                            int((1 - self.prob) * self.gaussian.shape[0] / self.prob),
                            self.gaussian.shape[1],
                        )
                    ),
                    self.gaussian,
                )
            )
        assert (
            self.bw_fs.ndim == 1
            and self.ampl_ratio.ndim == 2
            and self.bw_fs.shape[0] == self.ampl_ratio.shape[1]
        )
        self.inplace = inplace

        self._n_bw_choices = len(self.ampl_ratio)
        self._n_gn_choices = len(self.gaussian)

    def forward(
        self,
        sig: Tensor,
        label: Optional[Tensor],
        *extra_tensors: Sequence[Tensor],
        **kwargs: Any
    ) -> Tuple[Tensor, ...]:
        """Forward function of the :class:`BaselineWanderAugmenter`.

        Parameters
        ----------
        sig : torch.Tensor
            Batched ECGs to be augmented, of shape ``(batch, lead, siglen)``.
        label : torch.Tensor, optional
            Batched label tensor of the ECGs.
            Not used, but kept for consistency with other augmenters.
        extra_tensors : Sequence[torch.Tensor], optional,
            Not used, but kept for consistency with other augmenters.
        **kwargs : dict, optional
            Not used, but kept for consistency with other augmenters.

        Returns
        -------
        sig : torch.Tensor
            The augmented ECGs.
        label : torch.Tensor
            Label tensor of the augmented ECGs, unchanged.
        extra_tensors : Sequence[torch.Tensor], optional
            Unchanged extra tensors.

        """
        if not self.inplace:
            sig = sig.clone()
        if self.prob > 0:
            sig.add_(
                gen_baseline_wander(
                    sig, self.fs, self.bw_fs, self.ampl_ratio, self.gaussian
                )
            )
        return (sig, label, *extra_tensors)

    def extra_repr_keys(self) -> List[str]:
        return [
            "fs",
            "bw_fs",
            "prob",
            "inplace",
        ] + super().extra_repr_keys()


def _get_ampl(sig: Tensor, fs: int) -> Tensor:
    """Get the amplitude of each lead.

    Parameters
    ----------
    sig : torch.Tensor
        Batched ECG signal tensor, of shape ``(batch, lead, siglen)``.
    fs : int
        Sampling frequency of the ECGs.

    Returns
    -------
    ampl : torch.Tensor
        Amplitude of each lead, of shape ``(batch, lead, 1)``.

    """
    with mp.Pool(processes=max(1, mp.cpu_count() - 2)) as pool:
        ampl = pool.starmap(
            get_ampl,
            iterable=[(sig[i].cpu().numpy(), fs) for i in range(sig.shape[0])],
        )
    ampl = torch.as_tensor(
        np.array(ampl), dtype=sig.dtype, device=sig.device
    ).unsqueeze(-1)
    return ampl


def _gen_gaussian_noise(siglen: int, mean: Real = 0, std: Real = 0) -> np.ndarray:
    """Generate 1d Gaussian noise of given
    length, mean, and standard deviation.

    Parameters
    ----------
    siglen : int
        Length of the noise signal.
    mean : numbers.Real, default 0
        Mean value of the noise.
    std : numbers.Real, default 0
        Standard deviation of the noise.

    Returns
    -------
    gn : numpy.ndarray
        Gaussian noise of given length, mean, and standard deviation.

    """
    gn = DEFAULTS.RNG.normal(mean, std, siglen)
    return gn


def _gen_sinusoidal_noise(
    siglen: int,
    start_phase: Real,
    end_phase: Real,
    amplitude: Real,
    amplitude_mean: Real = 0,
    amplitude_std: Real = 0,
) -> np.ndarray:
    """Generate 1d sinusoidal noise of given
    length, amplitude, start phase, and end phase.

    Parameters
    ----------
    siglen : int
        Length of the (noise) signal.
    start_phase : numbers.Real
        Start phase, with units in degrees.
    end_phase : numbers.Real
        End phase, with units in degrees.
    amplitude : numbers.Real
        Amplitude of the sinusoidal curve.
    amplitude_mean : numbers.Real
        Mean amplitude of an extra Gaussian noise.
    amplitude_std : numbers.Real, default 0
        Standard deviation of an extra Gaussian noise

    Returns
    -------
    sn : numpy.ndarray,
        Sinusoidal noise of given length, amplitude, start phase, and end phase.

    """
    sn = np.linspace(start_phase, end_phase, siglen)
    sn = amplitude * np.sin(np.pi * sn / 180)
    sn += _gen_gaussian_noise(siglen, amplitude_mean, amplitude_std)
    return sn


def _gen_baseline_wander(
    siglen: int,
    fs: Real,
    bw_fs: Union[Real, Sequence[Real]],
    amplitude: Union[Real, Sequence[Real]],
    amplitude_gaussian: Sequence[Real] = [0, 0],
) -> np.ndarray:
    """Generate 1d baseline wander of given
    length, amplitude, and frequency.

    Parameters
    ----------
    siglen : int
        Length of the (noise) signal.
    fs : numbers.Real
        Sampling frequency of the original signal.
    bw_fs : numbers.Real, or list of numbers.Real
        Frequency (Frequencies) of the baseline wander.
    amplitude : numbers.Real, or list of numbers.Real
        Amplitude of the baseline wander (corr. to each frequency band).
    amplitude_gaussian : Tuple[numbers.Real], default [0,0]
        2-tuple of :class:`~numbers.Real`.
        Mean and std of amplitude of an extra Gaussian noise.

    Returns
    -------
    bw : numpy.ndarray
        Baseline wander of given length, amplitude, frequency.

    Example
    -------
    >>> _gen_baseline_wander(4000, 400, [0.4,0.1,0.05], [0.1,0.2,0.4])

    """
    bw = _gen_gaussian_noise(siglen, amplitude_gaussian[0], amplitude_gaussian[1])
    if isinstance(bw_fs, Real):
        _bw_fs = [bw_fs]
    else:
        _bw_fs = bw_fs
    if isinstance(amplitude, Real):
        _amplitude = list(repeat(amplitude, len(_bw_fs)))
    else:
        _amplitude = amplitude
    assert len(_bw_fs) == len(_amplitude)
    duration = siglen / fs
    for bf, a in zip(_bw_fs, _amplitude):
        start_phase = DEFAULTS.RNG_randint(0, 360)
        end_phase = duration * bf * 360 + start_phase
        bw += _gen_sinusoidal_noise(siglen, start_phase, end_phase, a, 0, 0)
    return bw


def gen_baseline_wander(
    sig: Tensor,
    fs: Real,
    bw_fs: Union[Real, Sequence[Real]],
    ampl_ratio: np.ndarray,
    gaussian: np.ndarray,
) -> np.ndarray:
    """Generate 1d baseline wander of given
    length, amplitude, and frequency.

    Parameters
    ----------
    sig : torch.Tensor
        Batched ECGs to be augmented, of shape (batch, lead, siglen).
    fs : numbers.Real
        Sampling frequency of the original signal.
    bw_fs : numbers.Real, or list of numbers.Real,
        Frequency (Frequencies) of the baseline wander.
    ampl_ratio : numpy.ndarray, optional
        Candidate ratios of noise amplitdes compared to the original ECGs for each `fs`,
        of shape ``(m, n)``.
    gaussian : numpy.ndarray, optional
        Candidate mean and std of the Gaussian noises,
        of shape ``(k, 2)``.

    Returns
    -------
    bw : numpy.ndarray
        Baseline wander of given length, amplitude, frequency,
        of shape ``(batch, lead, siglen)``.

    """
    batch, lead, siglen = sig.shape
    sig_ampl = _get_ampl(sig, fs)
    _n_bw_choices = len(ampl_ratio)
    _n_gn_choices = len(gaussian)

    with mp.Pool(processes=max(1, mp.cpu_count() - 2)) as pool:
        bw = pool.starmap(
            _gen_baseline_wander,
            iterable=[
                (
                    siglen,
                    fs,
                    bw_fs,
                    ampl_ratio[randint(0, _n_bw_choices - 1)],
                    gaussian[randint(0, _n_gn_choices - 1)],
                )
                for i in range(sig.shape[0])
                for j in range(sig.shape[1])
            ],
        )
    bw = torch.as_tensor(np.array(bw), dtype=sig.dtype, device=sig.device).reshape(
        batch, lead, siglen
    )
    return bw



"""
torch_ecg._preprocessors
========================

This module contains a set of preprocessors for signals of numpy array format.

.. contents:: torch_ecg._preprocessors
    :depth: 2
    :local:
    :backlinks: top

.. currentmodule:: torch_ecg._preprocessors

.. autosummary::
    :toctree: generated/
    :recursive:

    PreprocManager
    PreProcessor
    BandPass
    BaselineRemove
    Normalize
    MinMaxNormalize
    NaiveNormalize
    ZScoreNormalize
    Resample
    preprocess_multi_lead_signal
    preprocess_single_lead_signal

"""

from .bandpass import BandPass
from .base import (
    PreProcessor,
    preprocess_multi_lead_signal,
    preprocess_single_lead_signal,
)
from .baseline_remove import BaselineRemove
from .normalize import MinMaxNormalize, NaiveNormalize, Normalize, ZScoreNormalize
from .preproc_manager import PreprocManager
from .resample import Resample

__all__ = [
    "PreprocManager",
    "PreProcessor",
    "BandPass",
    "BaselineRemove",
    "Normalize",
    "MinMaxNormalize",
    "NaiveNormalize",
    "ZScoreNormalize",
    "Resample",
    "preprocess_multi_lead_signal",
    "preprocess_single_lead_signal",
]



"""
"""

import warnings
from typing import Any, Optional, Sequence, Union

import torch
import torch.nn as nn
from einops.layers.torch import Rearrange

from ...utils.utils_nn import SizeMixin
from ...utils.misc import get_kwargs


__all__ = [
    "Transformer",
]


class Transformer(nn.Module, SizeMixin):
    """Transformer feature extractor.

    Parameters
    ----------
    input_size : int
        Number of input channels.
    hidden_size : int
        Number of hidden units in the encoding layer.
    num_heads : int, default 8
        Number of attention heads.
    num_layers : int, default 1
        Number of encoding layers.
    dropout : float, default 0.1
        Dropout probability
    batch_first : bool, default True
        If True, the input is of shape ``(batch_size, seq_len, input_size)``,
        otherwise, the input is of shape ``(seq_len, batch_size, input_size)``.
    kwargs : dict
        Other hyper-parameters of the Module, including
        activation, kernel_initializer, etc.

    """

    __name__ = "Transformer"

    def __init__(
        self,
        input_size: int,
        hidden_size: int,
        num_heads: int = 8,
        num_layers: int = 1,
        dropout: float = 0.1,
        batch_first: bool = True,
        **kwargs: Any,
    ) -> None:
        super().__init__()
        self.__input_size = input_size
        self.__hidden_size = hidden_size
        self.__num_layers = num_layers
        self.__num_heads = num_heads
        self.__batch_first = batch_first

        if self.__input_size % self.__num_heads != 0:
            input_size = self.__input_size
            self.__input_size = self.__input_size // self.__num_heads * self.__num_heads
            warnings.warn(
                f"`input_size` {input_size} is not divisible by `num_heads` {self.__num_heads}, "
                f"adjusted to {self.__input_size}",
                RuntimeWarning,
            )
            if self.__batch_first:
                self.project = nn.Linear(input_size, self.__input_size)
            else:
                self.project = nn.Sequential(
                    Rearrange(
                        "seq_len batch_size input_size -> batch_size seq_len input_size"
                    ),
                    nn.Linear(input_size, self.__input_size),
                    Rearrange(
                        "batch_size seq_len input_size -> seq_len batch_size input_size"
                    ),
                )
        else:
            self.project = nn.Identity()

        if "batch_first" in get_kwargs(nn.TransformerEncoderLayer):
            encoder_layer = nn.TransformerEncoderLayer(
                d_model=self.__input_size,
                nhead=self.__num_heads,
                dim_feedforward=self.__hidden_size,
                dropout=dropout,
                batch_first=self.__batch_first,
                activation=kwargs.get("activation", "relu"),
            )
        else:
            encoder_layer = nn.TransformerEncoderLayer(
                d_model=self.__input_size,
                nhead=self.__num_heads,
                dim_feedforward=self.__hidden_size,
                dropout=dropout,
                activation=kwargs.get("activation", "relu"),
            )
            if self.__batch_first:
                warnings.warn(
                    "`batch_first` only supports torch >= 1.9, defaults to False",
                    RuntimeWarning,
                )
            self.__batch_first = False
        self.encoder = nn.TransformerEncoder(
            encoder_layer, num_layers=self.__num_layers
        )

    def forward(self, x: torch.Tensor) -> torch.Tensor:
        """Forward pass of the transformer feature extractor.

        Parameters
        ----------
        x : torch.Tensor
            Input tensor,
            of shape ``(seq_len, batch_size, input_size)`` if `batch_first` is False,
            else ``(batch_size, seq_len, input_size)``.

        Returns
        -------
        torch.Tensor
            Output tensor,
            of shape ``(seq_len, batch_size, output_size)`` if `batch_first` is False,
            else ``(batch_size, seq_len, output_size)``.

        """
        x = self.project(x)
        return self.encoder(x)

    def compute_output_shape(
        self, seq_len: Optional[int] = None, batch_size: Optional[int] = None
    ) -> Sequence[Union[int, None]]:
        """Compute the output shape of the transformer feature extractor.

        Parameters
        ----------
        seq_len : int, optional
            Length of the input tensor.
        batch_size : int, optional
            Batch size of the input tensor.

        Returns
        -------
        output_shape : sequence
            Output shape of the transformer feature extractor.

        """
        if self.__batch_first:
            return (batch_size, seq_len, self.__input_size)
        else:
            return (seq_len, batch_size, self.__input_size)

    @property
    def batch_first(self) -> bool:
        return self.__batch_first



# -*- coding: utf-8 -*-

import math
from numbers import Real
from pathlib import Path
from typing import Any, Dict, List, Optional, Sequence, Tuple, Union

import numpy as np
import pandas as pd
import scipy.signal as SS
from scipy.io import loadmat

from ...cfg import CFG, DEFAULTS
from ...utils.misc import add_docstring
from ...utils.utils_interval import get_optimal_covering
from ..base import DEFAULT_FIG_SIZE_PER_SEC, CPSCDataBase, DataBaseInfo


__all__ = [
    "CPSC2020",
    "compute_metrics",
]


_CPSC2020_INFO = DataBaseInfo(
    title="""
    The 3rd China Physiological Signal Challenge 2020:
    Searching for Premature Ventricular Contraction (PVC) and Supraventricular Premature Beat (SPB) from Long-term ECGs
    """,
    about="""
    1. training data consists of 10 single-lead ECG recordings collected from arrhythmia patients, each of the recording last for about 24 hours
    2. data and annotations are stored in v5 .mat files
    3. A02, A03, A08 are patient with atrial fibrillation
    4. sampling frequency = 400 Hz
    5. Detailed information:

        +-----+-----+-----------+-----------+-----------+-----------+---------------+
        | rec | ?AF | Length(h) | # N beats | # V beats | # S beats | # Total beats |
        +=====+=====+===========+===========+===========+===========+===============+
        | A01 | No  | 25.89     | 109,062   | 0         | 24        | 109,086       |
        +-----+-----+-----------+-----------+-----------+-----------+---------------+
        | A02 | Yes | 22.83     | 98,936    | 4,554     | 0         | 103,490       |
        +-----+-----+-----------+-----------+-----------+-----------+---------------+
        | A03 | Yes | 24.70     | 137,249   | 382       | 0         | 137,631       |
        +-----+-----+-----------+-----------+-----------+-----------+---------------+
        | A04 | No  | 24.51     | 77,812    | 19,024    | 3,466     | 100,302       |
        +-----+-----+-----------+-----------+-----------+-----------+---------------+
        | A05 | No  | 23.57     | 94,614    | 1         | 25        | 94,640        |
        +-----+-----+-----------+-----------+-----------+-----------+---------------+
        | A06 | No  | 24.59     | 77,621    | 0         | 6         | 77,627        |
        +-----+-----+-----------+-----------+-----------+-----------+---------------+
        | A07 | No  | 23.11     | 73,325    | 15,150    | 3,481     | 91,956        |
        +-----+-----+-----------+-----------+-----------+-----------+---------------+
        | A08 | Yes | 25.46     | 115,518   | 2,793     | 0         | 118,311       |
        +-----+-----+-----------+-----------+-----------+-----------+---------------+
        | A09 | No  | 25.84     | 88,229    | 2         | 1,462     | 89,693        |
        +-----+-----+-----------+-----------+-----------+-----------+---------------+
        | A10 | No  | 23.64     | 72,821    | 169       | 9,071     | 82,061        |
        +-----+-----+-----------+-----------+-----------+-----------+---------------+

    6. challenging factors for accurate detection of SPB and PVC:
       amplitude variation; morphological variation; noise
    7. Challenge official website [1]_.
    """,
    note="""
    1. the records can roughly be classified into 4 groups:

        +----+--------------------+
        | N  | A01, A03, A05, A06 |
        +----+--------------------+
        | V  | A02, A08           |
        +----+--------------------+
        | S  | A09, A10           |
        +----+--------------------+
        | VS | A04, A07           |
        +----+--------------------+

    2. as premature beats and atrial fibrillation can co-exists
       (via the following code, and data from CINC2020),
       the situation becomes more complicated.

       .. code-block:: python

            >>> from utils.scoring_aux_data import dx_cooccurrence_all
            >>> dx_cooccurrence_all.loc["AF", ["PAC","PVC","SVPB","VPB"]]
            PAC     20
            PVC     19
            SVPB     4
            VPB     20
            Name: AF, dtype: int64

       this could also be seen from this dataset, via the following code as an example:

       .. code-block:: python

            >>> from data_reader import CPSC2020Reader as CR
            >>> db_dir = "/media/cfs/wenhao71/data/CPSC2020/TrainingSet/"
            >>> dr = CR(db_dir)
            >>> rec = dr.all_records[1]
            >>> dr.plot(rec, sampfrom=0, sampto=4000, ticks_granularity=2)

    3. PVC and SPB can also co-exist, as illustrated via the following code (from CINC2020):

       .. code-block:: python

            >>> from utils.scoring_aux_data import dx_cooccurrence_all
            >>> dx_cooccurrence_all.loc[["PVC","VPB"], ["PAC","SVPB",]]
            PAC SVPB
            PVC 14 1
            VPB 27 0
            and also from the following code:
            >>> for rec in dr.all_records:
            >>>     ann = dr.load_ann(rec)
            >>>     spb = ann["SPB_indices"]
            >>>     pvc = ann["PVC_indices"]
            >>>     if len(np.diff(spb)) > 0:
            >>>         print(f"{rec}: min dist among SPB = {np.min(np.diff(spb))}")
            >>>     if len(np.diff(pvc)) > 0:
            >>>         print(f"{rec}: min dist among PVC = {np.min(np.diff(pvc))}")
            >>>     diff = [s-p for s,p in product(spb, pvc)]
            >>>     if len(diff) > 0:
            >>>         print(f"{rec}: min dist between SPB and PVC = {np.min(np.abs(diff))}")
            A01: min dist among SPB = 630
            A02: min dist among SPB = 696
            A02: min dist among PVC = 87
            A02: min dist between SPB and PVC = 562
            A03: min dist among SPB = 7044
            A03: min dist among PVC = 151
            A03: min dist between SPB and PVC = 3750
            A04: min dist among SPB = 175
            A04: min dist among PVC = 156
            A04: min dist between SPB and PVC = 178
            A05: min dist among SPB = 182
            A05: min dist between SPB and PVC = 22320
            A06: min dist among SPB = 455158
            A07: min dist among SPB = 603
            A07: min dist among PVC = 153
            A07: min dist between SPB and PVC = 257
            A08: min dist among SPB = 2903029
            A08: min dist among PVC = 106
            A08: min dist between SPB and PVC = 350
            A09: min dist among SPB = 180
            A09: min dist among PVC = 7719290
            A09: min dist between SPB and PVC = 1271
            A10: min dist among SPB = 148
            A10: min dist among PVC = 708
            A10: min dist between SPB and PVC = 177

    """,
    usage=[
        "ECG arrhythmia (PVC, SPB) detection",
    ],
    issues="""
    1. currently, using `xqrs` as qrs detector,
       a lot more (more than 1000) rpeaks would be detected for A02, A07, A08,
       which might be caused by motion artefacts (or AF?);
       a lot less (more than 1000) rpeaks would be detected for A04.
       numeric details are as follows:

            +-----+-----+-----------------+---------------+
            | rec | ?AF | # beats by xqrs | # Total beats |
            +=====+=====+=================+===============+
            | A01 | No  | 109,502         | 109,086       |
            +-----+-----+-----------------+---------------+
            | A02 | Yes | 119,562         | 103,490       |
            +-----+-----+-----------------+---------------+
            | A03 | Yes | 135,912         | 137,631       |
            +-----+-----+-----------------+---------------+
            | A04 | No  | 92,746          | 100,302       |
            +-----+-----+-----------------+---------------+
            | A05 | No  | 94,674          | 94,640        |
            +-----+-----+-----------------+---------------+
            | A06 | No  | 77,955          | 77,627        |
            +-----+-----+-----------------+---------------+
            | A07 | No  | 98,390          | 91,956        |
            +-----+-----+-----------------+---------------+
            | A08 | Yes | 126,908         | 118,311       |
            +-----+-----+-----------------+---------------+
            | A09 | No  | 89,972          | 89,693        |
            +-----+-----+-----------------+---------------+
            | A10 | No  | 83,509          | 82,061        |
            +-----+-----+-----------------+---------------+

    2. (fixed by an official update) A04 has duplicate "PVC_indices" (13534856,27147621,35141190 all appear twice):
       before correction of `load_ann`

       .. code-block:: python

            >>> from collections import Counter
            >>> db_dir = "/mnt/wenhao71/data/CPSC2020/TrainingSet/"
            >>> data_gen = CPSC2020Reader(db_dir=db_dir,working_dir=db_dir)
            >>> rec = 4
            >>> ann = data_gen.load_ann(rec)
            >>> Counter(ann["PVC_indices"]).most_common()[:4]
            [(13534856, 2), (27147621, 2), (35141190, 2), (848, 1)]

    3. when extracting morphological features using augmented rpeaks for A04,

       .. code-block:: python

            RuntimeWarning: invalid value encountered in double_scalars

       would raise for

       .. math::

            R\\_value = (R\\_value - y_min) / (y\\_max - y\\_min)

       and for

       .. math::

            y\\_values[n] = (y\\_values[n] - y\\_min) / (y\\_max - y\\_min).

       This is caused by the 13882273-th sample, which is contained in "PVC_indices",
       however, whether it is a PVC beat, or just motion artefact, is in doubt!

    """,
    references=[
        "http://2020.icbeb.org/CSPC2020",
    ],
    doi="10.1166/jmihi.2020.3289",
)


@add_docstring(_CPSC2020_INFO.format_database_docstring(), mode="prepend")
class CPSC2020(CPSCDataBase):
    """
    Parameters
    ----------
    db_dir : str or pathlib.Path, optional
        Storage path of the database.
        If not specified, data will be fetched from Physionet.
    working_dir : str, optional
        Working directory, to store intermediate files and log files.
    verbose : int, default 1
        Level of logging verbosity.
    kwargs : dict, optional
        Auxilliary key word arguments

    """

    def __init__(
        self,
        db_dir: Optional[Union[str, Path]] = None,
        working_dir: Optional[Union[str, Path]] = None,
        verbose: int = 1,
        **kwargs: Any,
    ) -> None:
        super().__init__(
            db_name="cpsc2020",
            db_dir=db_dir,
            working_dir=working_dir,
            verbose=verbose,
            **kwargs,
        )

        self.fs = 400
        self.spacing = 1000 / self.fs
        self.rec_ext = "mat"
        self.ann_ext = "mat"

        self._df_records = None
        self._all_records = None
        self._all_annotations = None
        self._ls_rec()

        self.subgroups = CFG(
            {
                "N": [
                    "A01",
                    "A03",
                    "A05",
                    "A06",
                ],
                "V": ["A02", "A08"],
                "S": ["A09", "A10"],
                "VS": ["A04", "A07"],
            }
        )

        self.palette = {
            "spb": "yellow",
            "pvc": "red",
        }

    def _ls_rec(self) -> None:
        """Find all records in the database directory
        and store them (path, metadata, etc.) in some private attributes.
        """
        self._df_records = pd.DataFrame()
        n_records = 10
        all_records = [f"A{i:02d}" for i in range(1, 1 + n_records)]
        self._df_records["path"] = [
            path
            for path in self.db_dir.rglob(f"*.{self.rec_ext}")
            if path.stem in all_records
        ]
        self._df_records["record"] = self._df_records["path"].apply(lambda x: x.stem)
        self._df_records.set_index("record", inplace=True)

        all_annotations = [f"R{i:02d}" for i in range(1, 1 + n_records)]
        df_ann = pd.DataFrame()
        df_ann["ann_path"] = [
            path
            for path in self.db_dir.rglob(f"*.{self.ann_ext}")
            if path.stem in all_annotations
        ]
        df_ann["record"] = df_ann["ann_path"].apply(lambda x: x.stem.replace("R", "A"))
        df_ann.set_index("record", inplace=True)
        # take the intersection by the index of `df_ann` and `self._df_records`
        self._df_records = self._df_records.join(df_ann, how="inner")

        if len(self._df_records) > 0:
            if self._subsample is not None:
                size = min(
                    len(self._df_records),
                    max(1, int(round(self._subsample * len(self._df_records)))),
                )
                self._df_records = self._df_records.sample(
                    n=size, random_state=DEFAULTS.SEED, replace=False
                )

        self._all_records = self._df_records.index.tolist()
        self._all_annotations = (
            self._df_records["ann_path"].apply(lambda x: x.stem).tolist()
        )

    @property
    def all_annotations(self):
        return self._all_annotations

    @property
    def all_references(self):
        return self._all_annotations

    def get_subject_id(self, rec: Union[int, str]) -> int:
        """Attach a unique subject ID to the record.

        Parameters
        ----------
        rec : str or int
            Record name or index of the record in :attr:`all_records`.

        Returns
        -------
        pid : int
            the ``subject_id`` corr. to `rec`.

        """
        if isinstance(rec, int):
            rec = self[rec]
        return int(f"20{int(rec[1:]):08d}")

    def get_absolute_path(
        self,
        rec: Union[str, int],
        extension: Optional[str] = None,
        ann: bool = False,
    ) -> Path:
        """Get the absolute path of the record `rec`.

        Parameters
        ----------
        rec : str or int
            Record name or index of the record in :attr:`all_records`.
        extension : str, optional
            Extension of the file.
        ann : bool, default False
            Whether to get the annotation file path or not.

        Returns
        -------
        abs_path : pathlib.Path
            Absolute path of the file.

        """
        if isinstance(rec, int):
            rec = self[rec]
        if extension is not None and not extension.startswith("."):
            extension = f".{extension}"
        col = "ann_path" if ann else "path"
        abs_path = self._df_records.loc[rec, col].with_suffix(extension)
        return abs_path

    def load_data(
        self,
        rec: Union[int, str],
        sampfrom: Optional[int] = None,
        sampto: Optional[int] = None,
        data_format: str = "channel_first",
        units: str = "mV",
        fs: Optional[Real] = None,
    ) -> np.ndarray:
        """Load the ECG data of the record `rec`.

        Parameters
        ----------
        rec : str or int
            Record name or index of the record in :attr:`all_records`.
        sampfrom : int, optional
            Start index of the data to be loaded.
        sampto : int, optional
            End index of the data to be loaded.
        data_format : str, default "channel_first"
            Format of the ECG data,
            "channel_last" (alias "lead_last"), or
            "channel_first" (alias "lead_first"), or
            "flat" (alias "plain").
        units : str or None, default "mV"
            Units of the output signal,
            can also be "μV" (with aliases "uV", "muV").
        fs : numbers.Real, optional
            Frequency of the output signal.
            if not None, the loaded data will be resampled to this frequency;
            if None, the loaded data will be returned as is.

        Returns
        -------
        data : numpy.ndarray
            The loaded ECG data.

        """
        rec_fp = self.get_absolute_path(rec, self.rec_ext)
        data = loadmat(str(rec_fp))["ecg"].astype(DEFAULTS.DTYPE.NP)
        sf, st = (sampfrom or 0), (sampto or len(data))
        data = data[sf:st]
        if fs is not None and fs != self.fs:
            data = SS.resample_poly(data, fs, self.fs, axis=0).astype(data.dtype)
        if data_format.lower() in ["channel_first", "lead_first"]:
            data = data.T
        elif data_format.lower() in ["flat", "plain"]:
            data = data.flatten()
        elif data_format.lower() not in ["channel_last", "lead_last"]:
            raise ValueError(f"Invalid `data_format`: {data_format}")
        if units.lower() in ["uv", "muv", "μv"]:
            data = (1000 * data).astype(int)
        elif units.lower() != "mv":
            raise ValueError(f"Invalid `units`: {units}")
        return data

    def load_ann(
        self,
        rec: Union[int, str],
        sampfrom: Optional[int] = None,
        sampto: Optional[int] = None,
    ) -> Dict[str, np.ndarray]:
        """Load the annotations of the record `rec`.

        Parameters
        ----------
        rec : str or int
            Record name or index of the record in :attr:`all_records`.
        sampfrom : int, optional
            Start index of the data to be loaded.
        sampto : int, optional
            End index of the data to be loaded.

        Returns
        -------
        ann : dict
            Annotation dictionary with items (:class:`~numpy.ndarray`)
            "SPB_indices" and "PVC_indices",
            which record the indices of SPBs and PVCs.

        """
        ann_fp = self.get_absolute_path(rec, self.ann_ext, ann=True)
        ann = loadmat(str(ann_fp))["ref"]
        sf, st = (sampfrom or 0), (sampto or np.inf)
        spb_indices = ann["S_ref"][0, 0].flatten().astype(int)
        # drop duplicates
        spb_indices = np.array(sorted(list(set(spb_indices))), dtype=int)
        spb_indices = spb_indices[np.where((spb_indices >= sf) & (spb_indices < st))[0]]
        pvc_indices = ann["V_ref"][0, 0].flatten().astype(int)
        # drop duplicates
        pvc_indices = np.array(sorted(list(set(pvc_indices))), dtype=int)
        pvc_indices = pvc_indices[np.where((pvc_indices >= sf) & (pvc_indices < st))[0]]
        ann = {
            "SPB_indices": spb_indices,
            "PVC_indices": pvc_indices,
        }
        return ann

    def train_test_split_rec(self, test_rec_num: int = 2) -> Dict[str, List[str]]:
        """Split the records into train set and test (val) set.

        Parameters
        ----------
        test_rec_num : int, default 2
            Number of records for the test (val) set.

        Returns
        -------
        split_res : dict
            Split result dictionary,
            with items "train", "test",
            both of which are lists of record names.

        """
        if test_rec_num == 1:
            test_records = DEFAULTS.RNG_sample(self.subgroups.VS, 1).tolist()
        elif test_rec_num == 2:
            test_records = (
                DEFAULTS.RNG_sample(self.subgroups.VS, 1).tolist()
                + DEFAULTS.RNG_sample(self.subgroups.N, 1).tolist()
            )
        elif test_rec_num == 3:
            test_records = (
                DEFAULTS.RNG_sample(self.subgroups.VS, 1).tolist()
                + DEFAULTS.RNG_sample(self.subgroups.N, 2).tolist()
            )
        elif test_rec_num == 4:
            test_records = []
            for k in self.subgroups.keys():
                test_records += DEFAULTS.RNG_sample(self.subgroups[k], 1).tolist()
        elif 5 <= test_rec_num <= 10:
            raise ValueError("test data ratio too high")
        else:
            raise ValueError("Invalid `test_rec_num`")
        train_records = [r for r in self.all_records if r not in test_records]

        split_res = CFG(
            {
                "train": train_records,
                "test": test_records,
            }
        )

        return split_res

    def locate_premature_beats(
        self,
        rec: Union[int, str],
        premature_type: Optional[str] = None,
        window: Real = 10,
        sampfrom: Optional[int] = None,
        sampto: Optional[int] = None,
    ) -> List[List[int]]:
        """Locate the sample indices of premature beats in a record.

        The locations are in the form of a list of lists, and
        each list contains the interval of sample indices of premature beats.

        Parameters
        ----------
        rec : str or int
            Record name or index of the record in :attr:`all_records`.
        premature_type : str, optional
            Premature beat type, can be one of "SPB", "PVC".
            If not specified, both SPBs and PVCs will be located.
        window : numbers.Real, default 10
            Window length of each premature beat,
            with units in seconds.
        sampfrom : int, optional
            Start index of the premature beats to locate.
        sampto : int, optional
            End index of the premature beats to locate.

        Returns
        -------
        premature_intervals : list
            List of intervals of premature beats.

        """
        ann = self.load_ann(rec)
        if premature_type:
            premature_inds = ann[f"{premature_type.upper()}_indices"]
        else:
            premature_inds = np.append(ann["SPB_indices"], ann["PVC_indices"])
            premature_inds = np.sort(premature_inds)
        try:  # premature_inds empty?
            sf, st = (sampfrom or 0), (sampto or premature_inds[-1] + 1)
        except Exception:
            premature_intervals = []
            return premature_intervals
        premature_inds = premature_inds[(sf < premature_inds) & (premature_inds < st)]
        tot_interval = [sf, st]
        windown_len = int(window * self.fs)
        premature_intervals = get_optimal_covering(
            total_interval=tot_interval,
            to_cover=premature_inds,
            min_len=windown_len,
            isolated_point_dist_threshold=windown_len // 2,
            split_threshold=windown_len,
            traceback=False,
        )
        return premature_intervals

    def plot(
        self,
        rec: Union[int, str],
        data: Optional[np.ndarray] = None,
        ann: Optional[Dict[str, np.ndarray]] = None,
        ticks_granularity: int = 0,
        sampfrom: Optional[int] = None,
        sampto: Optional[int] = None,
        rpeak_inds: Optional[Union[Sequence[int], np.ndarray]] = None,
    ) -> None:
        """Plot the ECG signal of a record.

        Parameters
        ----------
        rec : str or int
            Record name or index of the record in :attr:`all_records`.
        data : numpy.ndarray, optional
            ECG signal to plot.
            If given, data of `rec` will not be used.
            This is useful when plotting filtered data.
        ann : dict, optional
            Annotations for `data`, covering those from annotation files,
            with items "SPB_indices", "PVC_indices",
            each of which is a :class:`~numpy.ndarray`.
            Ignored if `data` is None.
        ticks_granularity : int, default 0
            Granularity to plot axis ticks, the higher the more ticks.
            0 (no ticks) --> 1 (major ticks) --> 2 (major + minor ticks)
        sampfrom : int, optional
            Start index of the data to plot.
        sampto : int, optional
            End index of the data to plot.
        rpeak_inds : array_like, optional
            Indices of R peaks.
            If `data` is None,
            then indices should be the absolute indices in the record.

        """
        if isinstance(rec, int):
            rec = self[rec]
        if "plt" not in dir():
            import matplotlib.pyplot as plt
        import matplotlib.patches as mpatches

        patches = {}

        if data is None:
            _data = self.load_data(
                rec, units="μV", sampfrom=sampfrom, sampto=sampto, data_format="flat"
            )
        else:
            units = self._auto_infer_units(data)
            if units == "mV":
                _data = data * 1000
            elif units == "μV":
                _data = data.copy()

        if ann is None or data is None:
            ann = self.load_ann(rec, sampfrom=sampfrom, sampto=sampto)
        sf, st = (sampfrom or 0), (sampto or len(_data))
        spb_indices = ann["SPB_indices"]
        pvc_indices = ann["PVC_indices"]
        spb_indices = spb_indices - sf
        pvc_indices = pvc_indices - sf

        if rpeak_inds is not None:
            if data is not None:
                rpeak_secs = np.array(rpeak_inds) / self.fs
            else:
                rpeak_secs = np.array(rpeak_inds)
                rpeak_secs = rpeak_secs[
                    np.where((rpeak_secs >= sf) & (rpeak_secs < st))[0]
                ]
                rpeak_secs = (rpeak_secs - sf) / self.fs

        line_len = self.fs * 25  # 25 seconds
        nb_lines = math.ceil(len(_data) / line_len)

        bias_thr = 0.15
        winL = 0.06
        winR = 0.08

        for idx in range(nb_lines):
            seg = _data[idx * line_len : (idx + 1) * line_len]
            secs = (np.arange(len(seg)) + idx * line_len) / self.fs
            fig_sz_w = int(round(DEFAULT_FIG_SIZE_PER_SEC * len(seg) / self.fs))
            y_range = np.max(np.abs(seg)) + 100
            fig_sz_h = 6 * y_range / 1500
            fig, ax = plt.subplots(figsize=(fig_sz_w, fig_sz_h))
            ax.plot(
                secs,
                seg,
                color="black",
                linewidth="2.0",
            )
            ax.axhline(y=0, linestyle="-", linewidth="1.0", color="red")
            if ticks_granularity >= 1:
                ax.xaxis.set_major_locator(plt.MultipleLocator(0.2))
                ax.yaxis.set_major_locator(plt.MultipleLocator(500))
                ax.grid(which="major", linestyle="-", linewidth="0.5", color="red")
            if ticks_granularity >= 2:
                ax.xaxis.set_minor_locator(plt.MultipleLocator(0.04))
                ax.yaxis.set_minor_locator(plt.MultipleLocator(100))
                ax.grid(which="minor", linestyle=":", linewidth="0.5", color="black")
            seg_spb = np.where(
                (spb_indices >= idx * line_len) & (spb_indices < (idx + 1) * line_len)
            )[0]
            # print(f"spb_indices = {spb_indices}, seg_spb = {seg_spb}")
            if len(seg_spb) > 0:
                seg_spb = spb_indices[seg_spb] / self.fs
                patches["SPB"] = mpatches.Patch(color=self.palette["spb"], label="SPB")
            seg_pvc = np.where(
                (pvc_indices >= idx * line_len) & (pvc_indices < (idx + 1) * line_len)
            )[0]
            # print(f"pvc_indices = {pvc_indices}, seg_pvc = {seg_pvc}")
            if len(seg_pvc) > 0:
                seg_pvc = pvc_indices[seg_pvc] / self.fs
                patches["PVC"] = mpatches.Patch(color=self.palette["pvc"], label="PVC")
            for t in seg_spb:
                ax.axvspan(
                    max(secs[0], t - bias_thr),
                    min(secs[-1], t + bias_thr),
                    color=self.palette["spb"],
                    alpha=0.3,
                )
                ax.axvspan(
                    max(secs[0], t - winL),
                    min(secs[-1], t + winR),
                    color=self.palette["spb"],
                    alpha=0.9,
                )
            for t in seg_pvc:
                ax.axvspan(
                    max(secs[0], t - bias_thr),
                    min(secs[-1], t + bias_thr),
                    color=self.palette["pvc"],
                    alpha=0.3,
                )
                ax.axvspan(
                    max(secs[0], t - winL),
                    min(secs[-1], t + winR),
                    color=self.palette["pvc"],
                    alpha=0.9,
                )
            if len(patches) > 0:
                ax.legend(
                    handles=[v for _, v in patches.items()],
                    loc="lower left",
                    prop={"size": 16},
                )
            if rpeak_inds is not None:
                seg_rpeak_secs = rpeak_secs[
                    np.where((rpeak_secs >= secs[0]) & (rpeak_secs < secs[-1]))[0]
                ]
                for r in seg_rpeak_secs:
                    ax.axvspan(r - 0.01, r + 0.01, color="green", alpha=0.7)
            ax.set_xlim(secs[0], secs[-1])
            ax.set_ylim(-y_range, y_range)
            ax.set_xlabel("Time [s]")
            ax.set_ylabel("Voltage [μV]")
            plt.show()

    @property
    def url(self) -> str:
        return (
            "https://opensz.oss-cn-beijing.aliyuncs.com/ICBEB2020/file/TrainingSet.zip"
        )

    @property
    def database_info(self) -> DataBaseInfo:
        return _CPSC2020_INFO

    @property
    def webpage(self) -> str:
        return "http://2020.icbeb.org/CSPC2020"


def compute_metrics(
    sbp_true: List[np.ndarray],
    pvc_true: List[np.ndarray],
    sbp_pred: List[np.ndarray],
    pvc_pred: List[np.ndarray],
    verbose: int = 0,
) -> Union[Tuple[int], dict]:
    """Score Function for all (test) records.

    Parameters
    ----------
    sbp_true, pvc_true, sbp_pred, pvc_pred : List[numpy.ndarray]
        List of numpy arrays of true and predicted SBP and PVC indices.
    verbose : int
        Verbosity level.

    Returns
    -------
    retval : tuple or dict
        Tuple of (negative) scores for each ectopic beat type (SBP, PVC),
        or dict of more scoring details, including

            - total_loss: sum of loss of each ectopic beat type (PVC and SPB)
            - true_positive: number of true positives of each ectopic beat type
            - false_positive: number of false positives of each ectopic beat type
            - false_negative: number of false negatives of each ectopic beat type

    """
    BaseCfg = CFG()
    BaseCfg.fs = 400
    BaseCfg.bias_thr = 0.15 * BaseCfg.fs
    s_score = np.zeros(
        [
            len(sbp_true),
        ],
        dtype=int,
    )
    v_score = np.zeros(
        [
            len(sbp_true),
        ],
        dtype=int,
    )
    # Scoring
    for i, (s_ref, v_ref, s_pos, v_pos) in enumerate(
        zip(sbp_true, pvc_true, sbp_pred, pvc_pred)
    ):
        s_tp = 0
        s_fp = 0
        s_fn = 0
        v_tp = 0
        v_fp = 0
        v_fn = 0
        # SBP
        if s_ref.size == 0:
            s_fp = len(s_pos)
        else:
            for m, ans in enumerate(s_ref):
                s_pos_cand = np.where(abs(s_pos - ans) <= BaseCfg.bias_thr)[0]
                if s_pos_cand.size == 0:
                    s_fn += 1
                else:
                    s_tp += 1
                    s_fp += len(s_pos_cand) - 1
        # PVC
        if v_ref.size == 0:
            v_fp = len(v_pos)
        else:
            for m, ans in enumerate(v_ref):
                v_pos_cand = np.where(abs(v_pos - ans) <= BaseCfg.bias_thr)[0]
                if v_pos_cand.size == 0:
                    v_fn += 1
                else:
                    v_tp += 1
                    v_fp += len(v_pos_cand) - 1
        # calculate the score
        s_score[i] = s_fp * (-1) + s_fn * (-5)
        v_score[i] = v_fp * (-1) + v_fn * (-5)

        if verbose >= 1:
            print(f"for the {i}-th record")
            print(f"s_tp = {s_tp}, s_fp = {s_fp}, s_fn = {s_fn}")
            print(f"v_tp = {v_tp}, v_fp = {v_fp}, s_fn = {v_fn}")
            print(f"s_score[{i}] = {s_score[i]}, v_score[{i}] = {v_score[i]}")

    Score1 = np.sum(s_score)
    Score2 = np.sum(v_score)

    if verbose >= 1:
        retval = CFG(
            total_loss=-(Score1 + Score2),
            class_loss={"S": -Score1, "V": -Score2},
            true_positive={"S": s_tp, "V": v_tp},
            false_positive={"S": s_fp, "V": v_fp},
            false_negative={"S": s_fn, "V": v_fn},
        )
    else:
        retval = Score1, Score2

    return retval



"""
WARNING
-------
This module is deprecated, please use `torch_ecg._preprocessors` instead.


preprocess of (single lead) ECG signal:
    band pass (dep. on purpose?) --> remove baseline (?) --> find rpeaks
    --> wave delineation (?, put into several stand alone files)

NOTE:
-----
1. ALL signals are assumed to have units in mV

References:
-----------
[1] https://github.com/PIA-Group/BioSPPy

"""

import multiprocessing as mp
from collections import Counter
from numbers import Real
from typing import Dict, List, Optional

import numpy as np

# from scipy.signal import medfilt
# https://github.com/scipy/scipy/issues/9680
from biosppy.signals.tools import filter_signal
from scipy.ndimage.filters import median_filter

from ..cfg import CFG
from .misc import list_sum, ms2samples
from .rpeaks import (
    christov_detect,
    engzee_detect,
    gamboa_detect,
    gqrs_detect,
    hamilton_detect,
    ssf_detect,
    xqrs_detect,
)


__all__ = [
    "preprocess_multi_lead_signal",
    "preprocess_single_lead_signal",
    "rpeaks_detect_multi_leads",
    "merge_rpeaks",
]


QRS_DETECTORS = {
    "xqrs": xqrs_detect,
    "gqrs": gqrs_detect,
    "hamilton": hamilton_detect,
    "ssf": ssf_detect,
    "christov": christov_detect,
    "engzee": engzee_detect,
    "gamboa": gamboa_detect,
}

# ECG signal preprocessing configurations
PreprocCfg = CFG()
# PreprocCfg.fs = 500
PreprocCfg.rpeak_mask_radius = 50  # ms
PreprocCfg.rpeak_lead_num_thr = (
    8 / 12
)  # ratio of leads, used for merging rpeaks detected from multiple leads
PreprocCfg.beat_winL = 250
PreprocCfg.beat_winR = 250


def preprocess_multi_lead_signal(
    raw_sig: np.ndarray,
    fs: Real,
    sig_fmt: str = "channel_first",
    bl_win: Optional[List[Real]] = None,
    band_fs: Optional[List[Real]] = None,
    rpeak_fn: Optional[str] = None,
    verbose: int = 0,
) -> Dict[str, np.ndarray]:
    """
    perform preprocessing for multi-lead ECG signal (with units in mV),
    preprocessing may include median filter, bandpass filter, and rpeaks detection, etc.

    Parameters
    ----------
    raw_sig: ndarray,
        the raw ECG signal, with units in mV
    fs: numbers.Real,
        sampling frequency of `raw_sig`
    sig_fmt: str, default "channel_first",
        format of the multi-lead ECG signal,
        "channel_last" (alias "lead_last"), or
        "channel_first" (alias "lead_first", original)
    bl_win: list (of 2 numbers.Real), optional,
        window (units in second) of baseline removal using `median_filter`,
        the first is the shorter one, the second the longer one,
        a typical pair is [0.2, 0.6],
        if is None or empty, baseline removal will not be performed
    band_fs: list (of 2 numbers.Real), optional,
        frequency band of the bandpass filter,
        a typical pair is [0.5, 45],
        be careful when detecting paced rhythm,
        if is None or empty, bandpass filtering will not be performed
    rpeak_fn: str, optional,
        name of the function detecting rpeaks,
        can be one of keys of `QRS_DETECTORS`, case insensitive
    verbose: int, default 0,
        print verbosity

    Returns
    -------
    retval: dict,
        with items
        - "filtered_ecg": the array of the processed ECG signal
        - "rpeaks": the array of indices of rpeaks; empty if `rpeak_fn` is not given

    """
    assert sig_fmt.lower() in [
        "channel_first",
        "lead_first",
        "channel_last",
        "lead_last",
    ]
    if sig_fmt.lower() in ["channel_last", "lead_last"]:
        filtered_ecg = raw_sig.T
    else:
        filtered_ecg = raw_sig.copy()
    rpeaks_candidates = []
    cpu_num = max(1, mp.cpu_count() - 3)
    with mp.Pool(processes=cpu_num) as pool:
        results = pool.starmap(
            func=preprocess_single_lead_signal,
            iterable=[
                (filtered_ecg[lead, ...], fs, bl_win, band_fs, rpeak_fn)
                for lead in range(filtered_ecg.shape[0])
            ],
        )
    for lead in range(filtered_ecg.shape[0]):
        filtered_metadata = results[lead]
        filtered_ecg[lead, ...] = filtered_metadata["filtered_ecg"]
        rpeaks_candidates.append(filtered_metadata["rpeaks"])
        if verbose >= 1:
            print(
                f"for the {lead}-th lead, rpeaks_candidates = {filtered_metadata['rpeaks']}"
            )

    if sig_fmt.lower() in ["channel_last", "lead_last"]:
        rpeaks = merge_rpeaks(rpeaks_candidates, raw_sig.T, fs, verbose)
    else:
        rpeaks = merge_rpeaks(rpeaks_candidates, raw_sig, fs, verbose)
    retval = CFG(
        {
            "filtered_ecg": filtered_ecg,
            "rpeaks": rpeaks,
        }
    )
    return retval


def preprocess_single_lead_signal(
    raw_sig: np.ndarray,
    fs: Real,
    bl_win: Optional[List[Real]] = None,
    band_fs: Optional[List[Real]] = None,
    rpeak_fn: Optional[str] = None,
    verbose: int = 0,
) -> Dict[str, np.ndarray]:
    """
    perform preprocessing for single lead ECG signal (with units in mV),
    preprocessing may include median filter, bandpass filter, and rpeaks detection, etc.

    Parameters
    ----------
    raw_sig: ndarray,
        the raw ECG signal, with units in mV
    fs: numbers.Real,
        sampling frequency of `raw_sig`
    bl_win: list (of 2 numbers.Real), optional,
        window (units in second) of baseline removal using `median_filter`,
        the first is the shorter one, the second the longer one,
        a typical pair is [0.2, 0.6],
        if is None or empty, baseline removal will not be performed
    band_fs: list (of 2 numbers.Real), optional,
        frequency band of the bandpass filter,
        a typical pair is [0.5, 45],
        be careful when detecting paced rhythm,
        if is None or empty, bandpass filtering will not be performed
    rpeak_fn: str, optional,
        name of the function detecting rpeaks,
        can be one of keys of `QRS_DETECTORS`, case insensitive
    verbose: int, default 0,
        print verbosity

    Returns
    -------
    retval: dict,
        with items
        - "filtered_ecg": the array of the processed ECG signal
        - "rpeaks": the array of indices of rpeaks; empty if `rpeak_fn` is not given

    """
    filtered_ecg = raw_sig.copy()

    # remove baseline
    if bl_win:
        window1 = 2 * (int(bl_win[0] * fs) // 2) + 1  # window size must be odd
        window2 = 2 * (int(bl_win[1] * fs) // 2) + 1
        baseline = median_filter(filtered_ecg, size=window1, mode="nearest")
        baseline = median_filter(baseline, size=window2, mode="nearest")
        filtered_ecg = filtered_ecg - baseline

    # filter signal
    if band_fs:
        filtered_ecg = filter_signal(
            signal=filtered_ecg,
            ftype="FIR",
            # ftype="butter",
            band="bandpass",
            order=int(0.3 * fs),
            sampling_rate=fs,
            frequency=band_fs,
        )["signal"]

    if rpeak_fn:
        rpeaks = QRS_DETECTORS[rpeak_fn.lower()](filtered_ecg, fs).astype(int)
    else:
        rpeaks = np.array([], dtype=int)

    retval = CFG(
        {
            "filtered_ecg": filtered_ecg,
            "rpeaks": rpeaks,
        }
    )

    return retval


def rpeaks_detect_multi_leads(
    sig: np.ndarray,
    fs: Real,
    sig_fmt: str = "channel_first",
    rpeak_fn: str = "xqrs",
    verbose: int = 0,
) -> np.ndarray:
    """
    detect rpeaks from the filtered multi-lead ECG signal (with units in mV)

    Parameters
    ----------
    sig: ndarray,
        the (better be filtered) ECG signal, with units in mV
    fs: numbers.Real,
        sampling frequency of `sig`
    sig_fmt: str, default "channel_first",
        format of the multi-lead ECG signal,
        "channel_last" (alias "lead_last"), or
        "channel_first" (alias "lead_first", original)
    rpeak_fn: str,
        name of the function detecting rpeaks,
        can be one of keys of `QRS_DETECTORS`, case insensitive
    verbose: int, default 0,
        print verbosity

    Returns
    -------
    rpeaks: np.ndarray,
        array of indices of the detected rpeaks of the multi-lead ECG signal

    """
    assert sig_fmt.lower() in [
        "channel_first",
        "lead_first",
        "channel_last",
        "lead_last",
    ]
    if sig_fmt.lower() in ["channel_last", "lead_last"]:
        s = sig.T
    else:
        s = sig.copy()
    rpeaks = []
    for lead in range(s.shape[0]):
        rpeaks.append((QRS_DETECTORS[rpeak_fn.lower()](s[lead], fs)).astype(int))
    rpeaks = merge_rpeaks(rpeaks, sig, fs, verbose)
    return rpeaks


def merge_rpeaks(
    rpeaks_candidates: List[np.ndarray], sig: np.ndarray, fs: Real, verbose: int = 0
) -> np.ndarray:
    """
    merge rpeaks that are detected from each lead of multi-lead signals (with units in mV),
    using certain criterion merging qrs masks from each lead

    Parameters
    ----------
    rpeaks_candidates: list of ndarray,
        each element (ndarray) is the array of indices of rpeaks of corr. lead
    sig: ndarray,
        the multi-lead ECG signal, with units in mV
    fs: numbers.Real,
        sampling frequency of `sig`
    verbose: int, default 0,
        print verbosity

    Returns
    -------
    final_rpeaks: np.ndarray
        the final rpeaks obtained by merging the rpeaks from all the leads

    """
    rpeak_masks = np.zeros_like(sig, dtype=int)
    sig_len = sig.shape[1]
    radius = ms2samples(PreprocCfg.rpeak_mask_radius, fs)
    if verbose >= 1:
        print(f"sig_len = {sig_len}, radius = {radius}")
    for lead in range(sig.shape[0]):
        for r in rpeaks_candidates[lead]:
            rpeak_masks[lead, max(0, r - radius) : min(sig_len - 1, r + radius)] = 1
    rpeak_masks = (
        rpeak_masks.sum(axis=0) >= int(PreprocCfg.rpeak_lead_num_thr * sig.shape[0])
    ).astype(int)
    rpeak_masks[0], rpeak_masks[-1] = 0, 0
    split_indices = np.where(np.diff(rpeak_masks) != 0)[0]
    if verbose >= 1:
        print(
            f"split_indices = {split_indices}, with total number = {len(split_indices)}"
        )
        if verbose >= 2:
            print(
                f"the corresponding intervals are {[[split_indices[2*idx], split_indices[2*idx+1]] for idx in range(len(split_indices)//2)]}"
            )

    final_rpeaks = []
    for idx in range(len(split_indices) // 2):
        start_idx = split_indices[2 * idx]
        end_idx = split_indices[2 * idx + 1]
        rc = list_sum(  # `lr`: list of rpeaks
            [
                lr[np.where((lr >= start_idx) & (lr <= end_idx))].tolist()
                for lr in rpeaks_candidates
            ]
        )
        if verbose >= 2:
            print(
                f"at the {idx}-th interval, start_idx = {start_idx}, end_idx = {end_idx}"
            )
            print(f"rpeak candidates = {rc}")
        counter = Counter(rc).most_common()
        if len(counter) > 0 and counter[0][1] >= len(rc) // 2 + 1:
            # might have the case where
            # rc is empty
            final_rpeaks.append(counter[0][0])
        elif len(counter) > 0:
            final_rpeaks.append(int(np.mean(rc)))
    final_rpeaks = np.array(final_rpeaks)
    return final_rpeaks



"""
"""

import itertools
from numbers import Real
from typing import Tuple

import numpy as np
import torch
import pytest

from torch_ecg.cfg import CFG, DEFAULTS
from torch_ecg._preprocessors import (
    PreProcessor,
    PreprocManager,
    BandPass,
    BaselineRemove,
    Normalize,
    MinMaxNormalize,
    NaiveNormalize,
    ZScoreNormalize,
    Resample,
    preprocess_multi_lead_signal,
    preprocess_single_lead_signal,
)


test_sig = torch.randn(12, 80000).numpy()


class DummyPreProcessor(PreProcessor):
    def __init__(self) -> None:
        super().__init__()

    def apply(self, sig: np.ndarray, fs: Real) -> Tuple[np.ndarray, int]:
        return sig, fs


def test_preproc_manager():
    sig = test_sig.copy()

    ppm = PreprocManager()
    assert ppm.empty
    sig, fs = ppm(sig, 200)
    ppm.add_(BandPass(0.5, 40))
    assert not ppm.empty
    ppm.add_(Resample(100), pos=0)
    ppm.add_(ZScoreNormalize())
    ppm.add_(BaselineRemove(), pos=1)
    assert len(ppm.preprocessors) == 4
    del ppm.preprocessors[-1]
    assert len(ppm.preprocessors) == 3
    ppm.add_(NaiveNormalize())
    ppm.add_(DummyPreProcessor(), pos=0)

    sig = test_sig.copy()
    sig, fs = ppm(sig, 200)

    config = CFG(
        random=True,
        resample={"fs": 500},
        bandpass={"filter_type": "fir"},
        normalize={"method": "min-max"},
        baseline_remove={"window1": 0.3, "window2": 0.7},
    )
    ppm = PreprocManager.from_config(config)
    assert ppm.random

    sig = test_sig.copy()
    sig, fs = ppm(sig, 200)

    ppm.random = False
    ppm.rearrange(
        new_ordering=[
            "baseline_remove",
            "resample",
            "bandpass",
            "normalize",
        ]
    )

    ppm.random = True
    with pytest.warns(
        RuntimeWarning, match="The preprocessors are applied in random order"
    ):
        ppm.rearrange(
            new_ordering=[
                "bandpass",
                "baseline_remove",
                "resample",
                "normalize",
            ]
        )
    ppm.random = False

    with pytest.raises(AssertionError, match="Duplicate preprocessor names"):
        ppm.rearrange(
            new_ordering=[
                "bandpass",
                "baseline_remove",
                "resample",
                "normalize",
                "bandpass",
            ]
        )
    with pytest.raises(AssertionError, match="Number of preprocessors mismatch"):
        ppm.rearrange(
            new_ordering=[
                "bandpass",
                "baseline_remove",
                "resample",
            ]
        )

    with pytest.raises(ValueError, match="Invalid input ECG signal"):
        ppm(DEFAULTS.RNG.normal(size=(1, 1, 1, 5000)), 200)

    config = {}
    with pytest.warns(
        RuntimeWarning,
        match="No preprocessors added to the manager\\. You are using a dummy preprocessor",
    ):
        ppm = PreprocManager.from_config(config)
    assert ppm.empty
    sig = test_sig.copy()
    sig, fs = ppm(sig, 200)

    assert str(ppm) == repr(ppm)

    del ppm, sig, fs


def test_bandpass():
    sig = test_sig.copy()
    bp = BandPass(0, 40)
    sig, fs = bp(sig, 200)
    bp = BandPass(0.5, None)
    sig, fs = bp(sig, 200)

    assert str(bp) == repr(bp)


def test_baseline_remove():
    sig = test_sig.copy()
    br = BaselineRemove()
    sig, fs = br(sig, 200)
    br = BaselineRemove(0.3, 0.9)
    sig, fs = br(sig, 200)

    with pytest.warns(
        RuntimeWarning, match="values of `window1` and `window2` are switched"
    ):
        br = BaselineRemove(0.9, 0.3)

    assert str(br) == repr(br)


def test_normalize():
    sig = test_sig.copy()
    std = 0.5 * np.ones(sig.shape[0])
    norm = Normalize(std=std, per_channel=True)
    sig, fs = norm(sig, 200)

    with pytest.raises(AssertionError, match="standard deviation should be positive"):
        norm = Normalize(std=0)
    with pytest.raises(
        AssertionError, match="standard deviations should all be positive"
    ):
        norm = Normalize(std=np.zeros(sig.shape[0]))

    assert str(norm) == repr(norm)

    norm = MinMaxNormalize(per_channel=True)
    sig, fs = norm(sig, 200)

    assert str(norm) == repr(norm)

    norm = NaiveNormalize(per_channel=True)
    sig, fs = norm(sig, 200)

    assert str(norm) == repr(norm)

    norm = ZScoreNormalize(per_channel=True)
    sig, fs = norm(sig, 200)

    assert str(norm) == repr(norm)


def test_resample():
    sig = test_sig.copy()
    rsmp = Resample(fs=500)
    sig, fs = rsmp(sig, 200)
    rsmp = Resample(siglen=5000)
    sig, fs = rsmp(sig, 200)

    with pytest.raises(
        AssertionError, match="one and only one of `fs` and `siglen` should be set"
    ):
        rsmp = Resample(fs=500, siglen=5000)
    with pytest.raises(
        AssertionError, match="one and only one of `fs` and `siglen` should be set"
    ):
        rsmp = Resample()

    assert str(rsmp) == repr(rsmp)


def test_preprocess_multi_lead_signal():
    sig = torch.randn(12, 8000).numpy()
    fs = 200

    grid = itertools.product(
        ["lead_first", "channel_last"],  # sig_fmt
        [None, [0.2, 0.6]],  # bl_win
        [None, [0.5, 45], [-np.inf, 40], [1, fs]],  # band_fs
        ["butter", "fir"],  # filter_type
    )
    for sig_fmt, bl_win, band_fs, filter_type in grid:
        if sig_fmt == "channel_last":
            filt_sig = sig.transpose(1, 0)
        else:
            filt_sig = sig.copy()
        filt_sig = preprocess_multi_lead_signal(
            filt_sig,
            fs,
            sig_fmt=sig_fmt,
            bl_win=bl_win,
            band_fs=band_fs,
            filter_type=filter_type,
        )

    with pytest.raises(
        AssertionError, match="multi-lead signal should be 2d or 3d array"
    ):
        preprocess_multi_lead_signal(sig[0], fs)
    with pytest.raises(
        AssertionError, match="multi-lead signal should be 2d or 3d array"
    ):
        preprocess_multi_lead_signal(torch.randn(1, 1, 12, 8000).numpy(), fs)

    with pytest.raises(
        AssertionError, match="multi-lead signal format `xxx` not supported"
    ):
        preprocess_multi_lead_signal(sig, fs, sig_fmt="xxx")

    with pytest.raises(AssertionError, match="Invalid frequency band"):
        preprocess_multi_lead_signal(sig, fs, band_fs=[1, 0.5])
    with pytest.raises(AssertionError, match="Invalid frequency band"):
        preprocess_multi_lead_signal(sig, fs, band_fs=[0, fs])

    with pytest.raises(ValueError, match="Unsupported filter type `xxx`"):
        preprocess_multi_lead_signal(sig, fs, band_fs=[0.5, 45], filter_type="xxx")


def test_preprocess_single_lead_signal():
    sig = torch.randn(8000).numpy()
    fs = 200

    grid = itertools.product(
        [None, [0.2, 0.6]],  # bl_win
        [None, [0.5, 45], [-np.inf, 40], [1, fs]],  # band_fs
        ["butter", "fir"],  # filter_type
    )
    for bl_win, band_fs, filter_type in grid:
        filt_sig = preprocess_single_lead_signal(
            sig,
            fs,
            bl_win=bl_win,
            band_fs=band_fs,
            filter_type=filter_type,
        )

    with pytest.raises(AssertionError, match="single-lead signal should be 1d array"):
        preprocess_single_lead_signal(sig[np.newaxis, ...], fs)

    with pytest.raises(AssertionError, match="Invalid frequency band"):
        preprocess_single_lead_signal(sig, fs, band_fs=[1, 0.5])
    with pytest.raises(AssertionError, match="Invalid frequency band"):
        preprocess_single_lead_signal(sig, fs, band_fs=[0, fs])

    with pytest.raises(ValueError, match="Unsupported filter type `xxx`"):
        preprocess_single_lead_signal(sig, fs, band_fs=[0.5, 45], filter_type="xxx")


def test_base_preprocessor():
    with pytest.raises(
        TypeError,
        match=f"Can't instantiate abstract class {PreProcessor.__name__} with abstract method",
    ):
        PreProcessor()



"""Manager of preprocessors."""

import warnings
from random import sample
from typing import List, Optional, Tuple

import numpy as np

from ..utils.misc import ReprMixin
from .bandpass import BandPass
from .base import PreProcessor
from .baseline_remove import BaselineRemove
from .normalize import Normalize
from .resample import Resample


__all__ = [
    "PreprocManager",
]


class PreprocManager(ReprMixin):
    """Manager of preprocessors.

    This class is used to manage a sequence of preprocessors. It can be used to
    add preprocessors to the manager, and apply the preprocessors to a signal.

    Parameters
    ----------
    pps : Tuple[PreProcessor], optional
        The sequence of preprocessors to be added to the manager.
    random : bool, default False
        Whether to apply the augmenters in random order.

    Examples
    --------
    .. code-block:: python

        import torch
        from torch_ecg.cfg import CFG
        from torch_ecg._preprocessors import PreprocManager
        config = CFG(
            random=False,
            resample={"fs": 500},
            bandpass={"filter_type": "fir"},
            normalize={"method": "min-max"},
        )
        ppm = PreprocManager.from_config(config)
        sig = torch.randn(12, 80000).numpy()
        sig, fs = ppm(sig, 200)

    """

    __name__ = "PreprocManager"

    def __init__(
        self, *pps: Optional[Tuple[PreProcessor, ...]], random: bool = False
    ) -> None:
        super().__init__()
        self.random = random
        self._preprocessors = list(pps)

    def _add_bandpass(self, **config: dict) -> None:
        """Add a bandpass preprocessor to the manager.

        Parameters
        ----------
        **config : dict
            The configuration of the bandpass preprocessor.

        """
        self._preprocessors.append(BandPass(**config))

    def _add_baseline_remove(self, **config: dict) -> None:
        """Add a baseline remove preprocessor to the manager.

        Parameters
        ----------
        **config : dict
            The configuration of the baseline remove preprocessor.

        """
        self._preprocessors.append(BaselineRemove(**config))

    def _add_normalize(self, **config: dict) -> None:
        """Add a normalize preprocessor to the manager.

        Parameters
        ----------
        **config : dict
            The configuration of the normalize preprocessor.

        """
        self._preprocessors.append(Normalize(**config))

    def _add_resample(self, **config: dict) -> None:
        """Add a resample preprocessor to the manager.

        Parameters
        ----------
        **config : dict
            The configuration of the resample preprocessor.

        """
        self._preprocessors.append(Resample(**config))

    def __call__(self, sig: np.ndarray, fs: int) -> Tuple[np.ndarray, int]:
        """The main function of the manager, which applies the preprocessors

        Parameters
        ----------
        sig : numpy.ndarray
            The signal to be preprocessed.
        fs : int
            Sampling frequency of the signal.

        Returns
        -------
        pp_sig : numpy.ndarray
            The preprocessed signal.
        new_fs : int
            Sampling frequency of the preprocessed signal.

        """
        if len(self.preprocessors) == 0:
            # raise ValueError("No preprocessors added to the manager.")
            # allow dummy (empty) preprocessors
            return sig, fs
        ordering = list(range(len(self.preprocessors)))
        if self.random:
            ordering = sample(ordering, len(ordering))
        pp_sig = sig
        new_fs = fs
        for idx in ordering:
            pp_sig, new_fs = self.preprocessors[idx](pp_sig, new_fs)
        return pp_sig, new_fs

    @classmethod
    def from_config(cls, config: dict) -> "PreprocManager":
        """Create a new instance of
        :class:`PreprocManager` from a configuration.

        Parameters
        ----------
        config : dict
            The configuration of the preprocessors,
            better to be an :class:`~collections.OrderedDict`.

        Returns
        -------
        ppm : PreprocManager
            A new instance of :class:`PreprocManager`.

        """
        ppm = cls(random=config.get("random", False))
        _mapping = {
            "bandpass": ppm._add_bandpass,
            "baseline_remove": ppm._add_baseline_remove,
            "normalize": ppm._add_normalize,
            "resample": ppm._add_resample,
        }
        for pp_name, pp_config in config.items():
            if pp_name in [
                "random",
                "fs",
            ]:
                continue
            if pp_name in _mapping and isinstance(pp_config, dict):
                _mapping[pp_name](**pp_config)
            else:
                # just ignore the other items
                pass
                # raise ValueError(f"Unknown preprocessor: {pp_name}")
        if ppm.empty:
            warnings.warn(
                "No preprocessors added to the manager. You are using a dummy preprocessor.",
                RuntimeWarning,
            )
        return ppm

    def rearrange(self, new_ordering: List[str]) -> None:
        """Rearrange the preprocessors.

        Parameters
        ----------
        new_ordering : List[str]
            The new ordering of the preprocessors.

        """
        if self.random:
            warnings.warn(
                "The preprocessors are applied in random order, "
                "rearranging the preprocessors will not take effect.",
                RuntimeWarning,
            )
        _mapping = {  # built-in preprocessors
            "Resample": "resample",
            "BandPass": "bandpass",
            "BaselineRemove": "baseline_remove",
            "Normalize": "normalize",
        }
        _mapping.update({v: k for k, v in _mapping.items()})
        for k in new_ordering:
            if k not in _mapping:
                # allow custom preprocessors
                assert k in [
                    item.__class__.__name__ for item in self._preprocessors
                ], f"Unknown preprocessor name: `{k}`"
                _mapping.update({k: k})
        assert len(new_ordering) == len(
            set(new_ordering)
        ), "Duplicate preprocessor names."
        assert len(new_ordering) == len(
            self._preprocessors
        ), "Number of preprocessors mismatch."
        self._preprocessors.sort(
            key=lambda item: new_ordering.index(_mapping[item.__class__.__name__])
        )

    def add_(self, pp: PreProcessor, pos: int = -1) -> None:
        """Add a (custom) preprocessor to the manager.

        This method is preferred against directly manipulating
        the internal list of preprocessors via
        :code:`PreprocManager.preprocessors.append(pp)`.

        Parameters
        ----------
        pp : PreProcessor
            The :class:`PreProcessor` to be added.
        pos : int, default -1
            The position to insert the preprocessor,
            should be >= -1, with -1 the indicator of the end.

        """
        assert isinstance(pp, PreProcessor)
        assert pp.__class__.__name__ not in [
            p.__class__.__name__ for p in self.preprocessors
        ], f"Preprocessor {pp.__class__.__name__} already exists."
        assert (
            isinstance(pos, int) and pos >= -1
        ), f"pos must be an integer >= -1, but got {pos}."
        if pos == -1:
            self._preprocessors.append(pp)
        else:
            self._preprocessors.insert(pos, pp)

    @property
    def preprocessors(self) -> List[PreProcessor]:
        return self._preprocessors

    @property
    def empty(self) -> bool:
        return len(self.preprocessors) == 0

    def extra_repr_keys(self) -> List[str]:
        return super().extra_repr_keys() + [
            "random",
            "preprocessors",
        ]
